from collections.abc import Callable
from typing import Any

import ehrax as rx
import equinox as eqx
import numpy as np
import pandas as pd
import pytest
import tables as tb
from ehrax.testing.common_setup import BINARY_OBSERVATION_CODE_INDEX, DATASET_SCHEME_MANAGER


def test_serialization_multi_subjects(tvx_ehr: rx.TVxEHR, tmpdir: str):
    path = f"{tmpdir}/tvx_ehr"
    tvx_ehr.save(path)
    loaded_ehr = rx.TVxEHR.load(path)
    assert tvx_ehr.equals(loaded_ehr)


class TestSampleSubjects:
    @pytest.fixture(params=[(1, 3), (111, 5)], scope="class")
    def sampled_tvx_ehr(self, tvx_ehr: rx.TVxEHR, request):
        seed, offset = request.param
        n_subjects = len(tvx_ehr.dataset.tables.static) // 5
        sample = rx.TVxEHRSampleConfig(seed=seed, n_subjects=n_subjects, offset=offset)
        tvx_ehr = eqx.tree_at(lambda x: x.config.sample, tvx_ehr, sample, is_leaf=lambda x: x is None)
        return rx.SampleSubjects.apply(tvx_ehr, DATASET_SCHEME_MANAGER, rx.TVxReport())[0]

    def test_sample_subjects(self, tvx_ehr: rx.TVxEHR, sampled_tvx_ehr: rx.TVxEHR):
        original_subjects = tvx_ehr.dataset.tables.static.index
        sampled_subjects = sampled_tvx_ehr.dataset.tables.static.index
        assert len(sampled_subjects) == len(original_subjects) // 5
        assert len(set(sampled_subjects)) == len(sampled_subjects)
        assert set(sampled_subjects).issubset(set(original_subjects))

    def test_ehr_serialization(self, tvx_ehr: rx.TVxEHR, sampled_tvx_ehr: rx.TVxEHR, tmpdir: str):
        path1 = f"{tmpdir}/tvx_ehr"
        path2 = f"{tmpdir}/sampled_tvx_ehr"

        tvx_ehr.save(path1)
        loaded_tvx_ehr = rx.TVxEHR.load(path1)

        sampled_tvx_ehr.save(path2)
        loaded_sampled_tvx_ehr = rx.TVxEHR.load(path2)

        assert not tvx_ehr.equals(sampled_tvx_ehr)
        assert not tvx_ehr.equals(loaded_sampled_tvx_ehr)
        assert not loaded_tvx_ehr.equals(sampled_tvx_ehr)
        assert not loaded_tvx_ehr.equals(loaded_sampled_tvx_ehr)
        assert sampled_tvx_ehr.equals(loaded_sampled_tvx_ehr)


class TestTrainableTransformer:
    @pytest.fixture(params=["obs", "icu_inputs"], scope="class")
    def scalable_table_name(self, request):
        return request.param

    @pytest.fixture(params=[True, False], scope="class")
    def use_float16(self, request):
        return request.param

    @pytest.fixture(scope="class")
    def scaler_class(self, scalable_table_name) -> type[rx.TrainableTransformation]:
        return {"obs": rx.ObsAdaptiveScaler, "icu_inputs": rx.InputScaler}[scalable_table_name]

    @pytest.fixture(params=[True, False], scope="class")
    def numerical_processor_config(self, request) -> rx.DatasetNumericalProcessorsConfig:
        null = request.param
        if null:
            return rx.DatasetNumericalProcessorsConfig()
        scalers_conf = rx.ScalersConfig(
            obs=rx.ScalerConfig(use_float16=True), icu_inputs=rx.ScalerConfig(use_float16=True)
        )
        outliers_conf = rx.OutlierRemoversConfig(obs=rx.IQROutlierRemoverConfig())
        return rx.DatasetNumericalProcessorsConfig(scalers_conf, outliers_conf)

    @pytest.fixture(scope="class")
    def large_scalable_split_ehr(self, tvx_ehr: rx.TVxEHR, scalable_table_name: str, use_float16: bool):
        if len(getattr(tvx_ehr.dataset.tables, scalable_table_name)) == 0:
            raise pytest.skip(f"No {scalable_table_name} table found in dataset.")
        subjects = tvx_ehr.dataset.tables.static.index.tolist()
        tvx_ehr = eqx.tree_at(lambda x: x.splits, tvx_ehr, (tuple(subjects),), is_leaf=lambda x: x is None)

        tvx_ehr = eqx.tree_at(
            lambda x: getattr(x.config.numerical_processors.scalers, scalable_table_name),
            tvx_ehr,
            rx.ScalerConfig(use_float16=use_float16),
            is_leaf=lambda x: x is None,
        )
        return tvx_ehr

    @pytest.fixture(scope="class")
    def scaled_ehr(self, large_scalable_split_ehr, scaler_class: type[rx.TrainableTransformation]):
        return scaler_class.apply(large_scalable_split_ehr, DATASET_SCHEME_MANAGER, rx.TVxReport())[0]

    def test_trainable_transformer(
        self, large_scalable_split_ehr: rx.TVxEHR, scaled_ehr: rx.TVxEHR, scalable_table_name: str, use_float16: bool
    ):
        assert getattr(large_scalable_split_ehr.numerical_processors.scalers, scalable_table_name) is None
        scaler = getattr(scaled_ehr.numerical_processors.scalers, scalable_table_name)
        assert isinstance(scaler, rx.CodedValueScaler)
        assert scaler.table_getter(scaled_ehr.dataset) is getattr(scaled_ehr.dataset.tables, scalable_table_name)
        assert scaler.table_getter(large_scalable_split_ehr.dataset) is getattr(
            large_scalable_split_ehr.dataset.tables, scalable_table_name
        )
        assert scaler.config.use_float16 == use_float16

        table0 = scaler.table_getter(large_scalable_split_ehr.dataset)
        table1 = scaler.table_getter(scaled_ehr.dataset)
        assert scaler.value_column in table1.columns
        assert scaler.code_column in table1.columns
        assert table1 is not table0
        if use_float16:
            assert table1[scaler.value_column].dtype == np.float16
        else:
            assert table1[scaler.value_column].dtype == table0[scaler.value_column].dtype

    @pytest.fixture(scope="class")
    def processed_ehr(
        self, large_scalable_split_ehr: rx.TVxEHR, numerical_processor_config: rx.DatasetNumericalProcessorsConfig
    ) -> rx.TVxEHR:
        large_scalable_split_ehr = eqx.tree_at(
            lambda x: x.config.numerical_processors, large_scalable_split_ehr, numerical_processor_config
        )
        return large_scalable_split_ehr._execute_pipeline(
            [rx.ObsIQROutlierRemover(), rx.InputScaler(), rx.ObsAdaptiveScaler()], DATASET_SCHEME_MANAGER
        )

    @pytest.fixture(scope="class")
    def fitted_numerical_processors(self, processed_ehr: rx.TVxEHR) -> rx.DatasetNumericalProcessors:
        return processed_ehr.numerical_processors

    def test_numerical_processors_serialization(
        self, fitted_numerical_processors: rx.DatasetNumericalProcessors, tmpdir: str
    ):
        path = f"{tmpdir}/numerical_processors.h5"
        with tb.open_file(path, "w") as f:
            fitted_numerical_processors.save(f.create_group("/", "numerical_processors"))

        with tb.open_file(path, "r") as f:
            loaded_numerical_processors = rx.DatasetNumericalProcessors.load(f.root.numerical_processors)
        assert fitted_numerical_processors.equals(loaded_numerical_processors)

    def test_ehr_serialization(
        self, tvx_ehr: rx.TVxEHR, large_scalable_split_ehr: rx.TVxEHR, processed_ehr: rx.TVxEHR, tmpdir: str
    ):
        assert not tvx_ehr.equals(large_scalable_split_ehr)
        assert not tvx_ehr.equals(processed_ehr)
        assert not large_scalable_split_ehr.equals(processed_ehr)

        split_path = f"{tmpdir}/split_ehr"
        processed_path = f"{tmpdir}/processed_ehr"
        large_scalable_split_ehr.save(split_path)
        loaded_split_ehr = rx.TVxEHR.load(split_path)
        assert large_scalable_split_ehr.equals(loaded_split_ehr)

        processed_ehr.save(processed_path)
        loaded_processed_ehr = rx.TVxEHR.load(processed_path)
        assert processed_ehr.equals(loaded_processed_ehr)


# def test_obs_minmax_scaler(int_dataset: Dataset):
#     assert False
#
#
# def test_obs_adaptive_scaler(int_dataset: Dataset):
#     assert False
#
#
# def test_obs_iqr_outlier_remover(dataset: Dataset):
#     assert False


# @pytest.mark.parametrize('splits', [[0.5], [0.2, 0.5, 0.7], [0.1, 0.2, 0.3, 0.4, 0.5]])
# def test_random_splits(dataset: Dataset, splits: list[float]):
# The logic of splits already tested in test.ehr.dataset.test_dataset.
# Maybe assert that functions are called with the correct arguments.
# pass


class TestTVxConcepts:
    @pytest.fixture(params=["gender", "age", "ethnicity"], scope="class")
    def tvx_ehr_demographic_config(self, request) -> rx.DemographicVectorConfig:
        config = rx.DemographicVectorConfig(False, False, False)
        return eqx.tree_at(lambda x: getattr(x, request.param), config, True)

    @pytest.fixture(scope="class")
    def tvx_ehr_configured_demographic(
        self, tvx_ehr: rx.TVxEHR, tvx_ehr_demographic_config: rx.DemographicVectorConfig
    ) -> rx.TVxEHR:
        return eqx.tree_at(lambda x: x.config.demographic, tvx_ehr, tvx_ehr_demographic_config)

    @pytest.fixture(scope="class")
    def tvx_concepts_static(self, tvx_ehr_configured_demographic: rx.TVxEHR) -> dict[str, rx.StaticInfo]:
        return rx.TVxConcepts._static_info(tvx_ehr_configured_demographic, DATASET_SCHEME_MANAGER, rx.TVxReport())[0]

    def test_tvx_concepts_static(
        self,
        tvx_ehr: rx.TVxEHR,
        tvx_concepts_static: dict[str, rx.StaticInfo],
        tvx_ehr_demographic_config: rx.DemographicVectorConfig,
    ):
        assert len(tvx_concepts_static) == len(tvx_ehr.dataset.tables.static)
        for subject_id, static_info in tvx_concepts_static.items():
            if tvx_ehr_demographic_config.gender:
                assert static_info.gender is not None
                assert len(static_info.gender.vec) == len(tvx_ehr.scheme_proxy(DATASET_SCHEME_MANAGER).gender)
                assert static_info.gender.vec.sum() == 1
                assert static_info.gender.scheme == tvx_ehr.scheme_proxy(DATASET_SCHEME_MANAGER).gender.name
            else:
                assert static_info.gender is None

            if tvx_ehr_demographic_config.age:
                assert static_info.date_of_birth is not None
                assert static_info.age(static_info.date_of_birth) == 0.0
            else:
                assert static_info.date_of_birth is None

            if tvx_ehr_demographic_config.ethnicity:
                assert static_info.ethnicity is not None
                assert len(static_info.ethnicity.vec) == len(tvx_ehr.scheme_proxy(DATASET_SCHEME_MANAGER).ethnicity)
                assert static_info.ethnicity.vec.sum() == 1
                assert static_info.ethnicity.scheme == tvx_ehr.scheme_proxy(DATASET_SCHEME_MANAGER).ethnicity.name

    @pytest.fixture(scope="class")
    def tvx_ehr_with_dx(self, tvx_ehr: rx.TVxEHR) -> rx.TVxEHR:
        if len(tvx_ehr.dataset.tables.dx_discharge) == 0:
            raise pytest.skip("No diagnoses table found in dataset.")
        n = len(tvx_ehr.dataset.tables.admissions)
        c_admission_id = tvx_ehr.dataset.config.columns.admissions.admission_id
        random_admission_id = tvx_ehr.dataset.tables.admissions.index[n // 2]
        dx_discharge = tvx_ehr.dataset.tables.dx_discharge
        dx_discharge = dx_discharge[dx_discharge[c_admission_id] != random_admission_id]
        return eqx.tree_at(lambda x: x.dataset.tables.dx_discharge, tvx_ehr, dx_discharge)

    @pytest.fixture(scope="class")
    def admission_dx_codes(self, tvx_ehr_with_dx: rx.TVxEHR) -> dict[str, rx.CodesVector]:
        return rx.TVxConcepts._dx_discharge(tvx_ehr_with_dx, DATASET_SCHEME_MANAGER)[0]

    @pytest.fixture(scope="class")
    def admission_dx_codeset(self, tvx_ehr_with_dx: rx.TVxEHR) -> dict[str, set[str]]:
        return rx.TVxConcepts._dx_discharge(tvx_ehr_with_dx, DATASET_SCHEME_MANAGER)[1]

    def test_admission_dx_codes(self, tvx_ehr_with_dx: rx.TVxEHR, admission_dx_codes: dict[str, rx.CodesVector]):
        assert set(admission_dx_codes.keys()) == set(tvx_ehr_with_dx.admission_ids)
        for admission_id, codes in admission_dx_codes.items():
            assert codes.vec.dtype == bool
            assert codes.scheme == tvx_ehr_with_dx.scheme_proxy(DATASET_SCHEME_MANAGER).dx_discharge.name
            assert len(codes.vec) == len(tvx_ehr_with_dx.scheme_proxy(DATASET_SCHEME_MANAGER).dx_discharge)

    @pytest.fixture(scope="class")
    def admission_dx_history_codes(
        self, tvx_ehr_with_dx: rx.TVxEHR, admission_dx_codes: dict[str, rx.CodesVector]
    ) -> dict[str, rx.CodesVector]:
        return rx.TVxConcepts._dx_discharge_history(tvx_ehr_with_dx, DATASET_SCHEME_MANAGER, admission_dx_codes)

    def test_admission_dx_history_codes(
        self,
        tvx_ehr_with_dx: rx.TVxEHR,
        admission_dx_codes: dict[str, rx.CodesVector],
        admission_dx_history_codes: dict[str, rx.CodesVector],
    ):
        assert set(admission_dx_history_codes.keys()) == set(tvx_ehr_with_dx.admission_ids)
        assert set(admission_dx_history_codes.keys()) == set(admission_dx_codes.keys())
        for subject_id, admission_ids in tvx_ehr_with_dx.subjects_sorted_admission_ids.items():
            assert admission_dx_history_codes[admission_ids[0]].vec.sum() == 0

            accumulation = np.zeros(len(tvx_ehr_with_dx.scheme_proxy(DATASET_SCHEME_MANAGER).dx_discharge), dtype=bool)
            for i, admission_id in enumerate(admission_ids):
                assert admission_id in admission_dx_history_codes
                history = admission_dx_history_codes[admission_id]
                assert history.vec.dtype == bool
                assert history.scheme == tvx_ehr_with_dx.scheme_proxy(DATASET_SCHEME_MANAGER).dx_discharge.name
                assert len(history.vec) == len(tvx_ehr_with_dx.scheme_proxy(DATASET_SCHEME_MANAGER).dx_discharge)
                assert (history.vec == accumulation).all()
                if admission_id in admission_dx_codes:
                    accumulation |= admission_dx_codes[admission_id].vec

    @pytest.fixture(scope="class")
    def admission_outcome(
        self, tvx_ehr_with_dx: rx.TVxEHR, admission_dx_codeset: dict[str, set[str]]
    ) -> dict[str, rx.CodesVector]:
        return rx.TVxConcepts._outcome(tvx_ehr_with_dx, DATASET_SCHEME_MANAGER, admission_dx_codeset)

    def test_admission_outcome(
        self,
        tvx_ehr_with_dx: rx.TVxEHR,
        admission_dx_codes: dict[str, rx.CodesVector],
        admission_outcome: dict[str, rx.CodesVector],
    ):
        assert set(admission_outcome.keys()) == set(admission_dx_codes.keys())
        outcome_scheme = tvx_ehr_with_dx.scheme_proxy(DATASET_SCHEME_MANAGER).outcome
        for admission_id, outcome in admission_outcome.items():
            assert outcome.scheme == outcome_scheme.name
            assert len(outcome.vec) == len(outcome_scheme)

    @pytest.fixture(scope="class")
    def tvx_ehr_with_icu_inputs(self, tvx_ehr: rx.TVxEHR) -> rx.TVxEHR:
        if len(tvx_ehr.dataset.tables.icu_inputs) == 0:
            raise pytest.skip("No icu_inputs table found in dataset.")
        return tvx_ehr

    @pytest.fixture(scope="class")
    def admission_icu_inputs(self, tvx_ehr_with_icu_inputs: rx.TVxEHR) -> dict[str, rx.InpatientInput]:
        return rx.TVxConcepts._icu_inputs(tvx_ehr_with_icu_inputs, DATASET_SCHEME_MANAGER)

    def test_admission_icu_inputs(
        self, tvx_ehr_with_icu_inputs: rx.TVxEHR, admission_icu_inputs: dict[str, rx.InpatientInput]
    ):
        icu_inputs = tvx_ehr_with_icu_inputs.dataset.tables.icu_inputs
        c_admission_id = tvx_ehr_with_icu_inputs.dataset.config.columns.admissions.admission_id
        assert set(admission_icu_inputs.keys()).issubset(set(tvx_ehr_with_icu_inputs.admission_ids))
        assert sum(len(inputs.starttime) for inputs in admission_icu_inputs.values()) == len(icu_inputs)

        for admission_id, admission_inputs_df in icu_inputs.groupby(c_admission_id):
            tvx_inputs = admission_icu_inputs[admission_id]
            assert len(tvx_inputs.starttime) == len(admission_inputs_df)
            assert all(
                tvx_inputs.code_index
                < len(tvx_ehr_with_icu_inputs.dataset.scheme_proxy(DATASET_SCHEME_MANAGER).icu_inputs)
            )

    @pytest.fixture(scope="class")
    def tvx_ehr_with_obs(self, tvx_ehr: rx.TVxEHR) -> rx.TVxEHR:
        if len(tvx_ehr.dataset.tables.obs) == 0:
            raise pytest.skip("No observations table found in dataset.")
        return tvx_ehr

    @pytest.fixture(scope="class")
    def admission_obs(self, tvx_ehr_with_obs: rx.TVxEHR) -> dict[str, rx.InpatientObservables]:
        return rx.TVxConcepts._observables(tvx_ehr_with_obs, DATASET_SCHEME_MANAGER, rx.TVxReport())[0]

    def test_admission_obs(self, tvx_ehr_with_obs: rx.TVxEHR, admission_obs: dict[str, rx.InpatientObservables]):
        obs_df = tvx_ehr_with_obs.dataset.tables.obs
        c_admission_id = tvx_ehr_with_obs.dataset.config.columns.admissions.admission_id
        assert set(admission_obs.keys()) == set(tvx_ehr_with_obs.admission_ids)
        assert sum(obs.mask.sum() for obs in admission_obs.values()) == len(obs_df)

        for admission_id, admission_obs_df in obs_df.groupby(c_admission_id):
            tvx_obs = admission_obs[admission_id]
            vals = tvx_obs.value.flatten()[tvx_obs.mask.flatten()]
            tvx_time_vals = pd.Series(tvx_obs.time)
            tbl_time_vals = pd.Series(np.unique(admission_obs_df[rx.COLUMN.time]))

            tvx_vals_count = pd.Series(vals, name=rx.COLUMN.measurement).value_counts().sort_index()
            tbl_vals_count = admission_obs_df[rx.COLUMN.measurement].astype(vals.dtype).value_counts().sort_index()

            assert tvx_vals_count.equals(tbl_vals_count)
            assert tvx_time_vals.equals(tbl_time_vals)
            assert tvx_obs.mask.sum() == len(admission_obs_df)
            assert tvx_obs.value.shape[1] == len(tvx_ehr_with_obs.dataset.scheme_proxy(DATASET_SCHEME_MANAGER).obs)

    @pytest.fixture(scope="class")
    def tvx_ehr_with_hosp_procedures(self, tvx_ehr: rx.TVxEHR) -> rx.TVxEHR:
        if len(tvx_ehr.dataset.tables.hosp_procedures) == 0:
            raise pytest.skip("No hospital procedures table found in dataset.")
        return tvx_ehr

    @pytest.fixture(scope="class")
    def admission_hosp_procedures(self, tvx_ehr_with_hosp_procedures: rx.TVxEHR) -> dict[str, rx.InpatientInput]:
        return rx.TVxConcepts._hosp_procedures(tvx_ehr_with_hosp_procedures, DATASET_SCHEME_MANAGER)

    def test_admission_hosp_procedures(
        self, tvx_ehr_with_hosp_procedures: rx.TVxEHR, admission_hosp_procedures: dict[str, rx.InpatientInput]
    ):
        hosp_procedures = tvx_ehr_with_hosp_procedures.dataset.tables.hosp_procedures
        c_admission_id = tvx_ehr_with_hosp_procedures.dataset.config.columns.admissions.admission_id
        assert set(admission_hosp_procedures.keys()).issubset(set(tvx_ehr_with_hosp_procedures.admission_ids))
        assert sum(len(proc.starttime) for proc in admission_hosp_procedures.values() if proc is not None) == len(
            hosp_procedures
        )

        for admission_id, admission_hosp_procedures_df in hosp_procedures.groupby(c_admission_id):
            tvx_hosp_proc = admission_hosp_procedures[admission_id]
            assert len(tvx_hosp_proc.starttime) == len(admission_hosp_procedures_df)
            assert all(
                tvx_hosp_proc.code_index
                < len(tvx_ehr_with_hosp_procedures.scheme_proxy(DATASET_SCHEME_MANAGER).hosp_procedures)
            )

    @pytest.fixture(scope="class")
    def tvx_ehr_with_icu_procedures(self, tvx_ehr: rx.TVxEHR) -> rx.TVxEHR:
        if len(tvx_ehr.dataset.tables.icu_procedures) == 0:
            raise pytest.skip("No icu procedures table found in dataset.")
        return tvx_ehr

    @pytest.fixture(scope="class")
    def admission_icu_procedures(self, tvx_ehr_with_icu_procedures: rx.TVxEHR) -> dict[str, rx.InpatientInput]:
        return rx.TVxConcepts._icu_procedures(tvx_ehr_with_icu_procedures, DATASET_SCHEME_MANAGER)

    def test_admission_icu_procedures(
        self, tvx_ehr_with_icu_procedures: rx.TVxEHR, admission_icu_procedures: dict[str, rx.InpatientInput]
    ):
        icu_procedures = tvx_ehr_with_icu_procedures.dataset.tables.icu_procedures
        c_admission_id = tvx_ehr_with_icu_procedures.dataset.config.columns.admissions.admission_id
        assert set(admission_icu_procedures.keys()).issubset(set(tvx_ehr_with_icu_procedures.admission_ids))
        assert sum(len(proc.starttime) for proc in admission_icu_procedures.values() if proc is not None) == len(
            icu_procedures
        )

        for admission_id, admission_icu_procedures_df in icu_procedures.groupby(c_admission_id):
            tvx_icu_proc = admission_icu_procedures[str(admission_id)]
            assert len(tvx_icu_proc.starttime) == len(admission_icu_procedures_df)
            assert all(
                tvx_icu_proc.code_index
                < len(tvx_ehr_with_icu_procedures.scheme_proxy(DATASET_SCHEME_MANAGER).icu_procedures)
            )

    @pytest.fixture(scope="class")
    def tvx_ehr_with_all_interventions(self, tvx_ehr: rx.TVxEHR) -> rx.TVxEHR:
        if tvx_ehr.dataset.tables.icu_procedures is None or len(tvx_ehr.dataset.tables.icu_procedures) == 0:
            raise pytest.skip("No icu procedures table found in dataset.")
        if tvx_ehr.dataset.tables.hosp_procedures is None or len(tvx_ehr.dataset.tables.hosp_procedures) == 0:
            raise pytest.skip("No hospital procedures table found in dataset.")
        if tvx_ehr.dataset.tables.icu_inputs is None or len(tvx_ehr.dataset.tables.icu_inputs) == 0:
            raise pytest.skip("No icu inputs table found in dataset.")
        return tvx_ehr

    @pytest.fixture(scope="class")
    def admission_interventions(
        self, tvx_ehr_with_all_interventions: rx.TVxEHR
    ) -> dict[str, rx.InpatientInterventions]:
        return rx.TVxConcepts._interventions(tvx_ehr_with_all_interventions, DATASET_SCHEME_MANAGER, rx.TVxReport())[0]

    def test_admission_interventions(
        self, tvx_ehr_with_all_interventions: rx.TVxEHR, admission_interventions: dict[str, rx.InpatientInterventions]
    ):
        assert set(admission_interventions.keys()) == set(tvx_ehr_with_all_interventions.admission_ids)
        for attr in ("icu_procedures", "hosp_procedures", "icu_inputs"):
            table = getattr(tvx_ehr_with_all_interventions.dataset.tables, attr)
            interventions = {admission_id: getattr(v, attr) for admission_id, v in admission_interventions.items()}
            assert sum(len(v.starttime) for v in interventions.values() if v is not None) == len(table)

    @pytest.fixture(params=["interventions", "observables"], scope="class")
    def tvx_ehr_conf_concept(self, tvx_ehr: rx.TVxEHR, request) -> rx.TVxEHR:
        concept_name = request.param
        conf = tvx_ehr.config
        for cname in ("interventions", "observables"):
            conf = eqx.tree_at(lambda x: getattr(x, cname), conf, concept_name == cname)
        return eqx.tree_at(lambda x: x.config, tvx_ehr, conf)

    @pytest.fixture(scope="class")
    def tvx_ehr_concept(self, tvx_ehr_conf_concept: rx.TVxEHR):
        return tvx_ehr_conf_concept._execute_pipeline([rx.TVxConcepts()], DATASET_SCHEME_MANAGER)

    def test_tvx_ehr_concept(self, tvx_ehr_concept: rx.TVxEHR):
        assert set(tvx_ehr_concept.subjects.keys()) == set(tvx_ehr_concept.dataset.tables.static.index)
        if tvx_ehr_concept.config.interventions:
            assert all(
                adm.interventions is not None
                for patient in tvx_ehr_concept.subjects.values()
                for adm in patient.admissions
            )
        else:
            assert all(
                adm.interventions is None for patient in tvx_ehr_concept.subjects.values() for adm in patient.admissions
            )
        if tvx_ehr_concept.config.observables:
            assert all(
                adm.observables is not None
                for patient in tvx_ehr_concept.subjects.values()
                for adm in patient.admissions
            )
        else:
            assert all(
                adm.observables is None for patient in tvx_ehr_concept.subjects.values() for adm in patient.admissions
            )

        for subject_id, sorted_admission_ids in tvx_ehr_concept.subjects_sorted_admission_ids.items():
            assert [adm.admission_id for adm in tvx_ehr_concept.subjects[subject_id].admissions] == sorted_admission_ids

    @pytest.fixture(scope="class")
    def mutated_ehr_concept(self, tvx_ehr_concept: rx.TVxEHR, tvx_ehr_conf_concept: rx.TVxEHR):
        if tvx_ehr_conf_concept.config.interventions:
            s0 = tvx_ehr_concept.subject_ids[0]
            a0 = tvx_ehr_concept.subjects[s0].admissions[0]
            a1 = eqx.tree_at(lambda x: x.interventions.icu_inputs.rate, a0, a0.interventions.icu_inputs.rate + 0.1)
            return eqx.tree_at(lambda x: x.subjects[s0].admissions[0], tvx_ehr_concept, a1)
        else:
            s0 = tvx_ehr_concept.subject_ids[0]
            a0 = tvx_ehr_concept.subjects[s0].admissions[0]
            a1 = eqx.tree_at(lambda x: x.observables.value, a0, a0.observables.value + 0.1)
            return eqx.tree_at(lambda x: x.subjects[s0].admissions[0], tvx_ehr_concept, a1)

    def test_ehr_serialization(
        self, tvx_ehr_conf_concept: rx.TVxEHR, tvx_ehr_concept: rx.TVxEHR, mutated_ehr_concept: rx.TVxEHR, tmpdir: str
    ):
        ehr_list = [tvx_ehr_conf_concept, tvx_ehr_concept, mutated_ehr_concept]
        for i, ehr_i in enumerate(ehr_list):
            path = f"{tmpdir}/tvx_ehr_{i}"
            ehr_i.save(path)
            loaded_ehr_i = rx.TVxEHR.load(path)
            for j in range(i + 1, len(ehr_list)):
                ehr_j = ehr_list[j]
                assert loaded_ehr_i.equals(ehr_j) == (i == j)


class TestInterventionSegmentation:
    @pytest.fixture(scope="class")
    def tvx_ehr_concept(self, tvx_ehr: rx.TVxEHR):
        tvx_ehr = eqx.tree_at(lambda x: x.config.interventions, tvx_ehr, True)
        tvx_ehr = eqx.tree_at(lambda x: x.config.observables, tvx_ehr, True)
        return tvx_ehr._execute_pipeline([rx.TVxConcepts()], DATASET_SCHEME_MANAGER)

    @pytest.fixture(scope="class")
    def tvx_ehr_segmented(self, tvx_ehr_concept: rx.TVxEHR) -> rx.TVxEHR:
        tvx_ehr_concept = eqx.tree_at(lambda x: x.config.interventions_segmentation, tvx_ehr_concept, True)
        return tvx_ehr_concept._execute_pipeline([rx.InterventionSegmentation()], DATASET_SCHEME_MANAGER)

    def test_segmentation(self, tvx_ehr_concept: rx.TVxEHR, tvx_ehr_segmented: rx.TVxEHR):
        first_patient = next(iter(tvx_ehr_concept.subjects.values()))
        first_segmented_patient = next(iter(tvx_ehr_segmented.subjects.values()))
        assert isinstance(first_patient, rx.Patient)
        assert isinstance(first_segmented_patient, rx.SegmentedPatient)
        assert isinstance(first_patient.admissions[0], rx.Admission)
        assert isinstance(first_segmented_patient.admissions[0], rx.SegmentedAdmission)
        assert isinstance(first_patient.admissions[0].observables, rx.InpatientObservables)
        assert isinstance(first_segmented_patient.admissions[0].observables, rx.SegmentedInpatientObservables)
        assert isinstance(first_segmented_patient.admissions[0].observables[0], rx.InpatientObservables)
        assert isinstance(first_patient.admissions[0].interventions, rx.InpatientInterventions)
        assert isinstance(first_segmented_patient.admissions[0].interventions, rx.SegmentedInpatientInterventions)

    @pytest.fixture(
        params=["mutate_obs", "intervention_time", "mutate_icu_proc", "mutate_hosp_proc", "mutate_icu_input"],
        scope="class",
    )
    def mutated_ehr_concept(self, tvx_ehr_segmented: rx.TVxEHR, request):
        s0 = tvx_ehr_segmented.subject_ids[0]
        a0 = tvx_ehr_segmented.subjects[s0].admissions[0]
        if request.param == "mutate_obs":
            a1 = eqx.tree_at(lambda x: x.observables.time, a0, a0.observables.time + 0.001)
        elif request.param == "intervention_time":
            a1 = eqx.tree_at(lambda x: x.interventions.time, a0, a0.interventions.time + 1e-6)
        elif request.param == "mutate_icu_proc":
            if a0.interventions.icu_procedures is None:
                raise pytest.skip("No icu procedures in admission.")
            a1 = eqx.tree_at(lambda x: x.interventions.icu_procedures, a0, ~a0.interventions.icu_procedures)

        elif request.param == "mutate_hosp_proc":
            if a0.interventions.hosp_procedures is None:
                raise pytest.skip("No hospital procedures in admission.")
            a1 = eqx.tree_at(lambda x: x.interventions.hosp_procedures, a0, ~a0.interventions.hosp_procedures)
        elif request.param == "mutate_icu_input":
            if a0.interventions.icu_inputs is None:
                raise pytest.skip("No icu inputs in admission.")
            assert a0.interventions.icu_inputs is not None
            a1 = eqx.tree_at(lambda x: x.interventions.icu_inputs, a0, a0.interventions.icu_inputs + 0.1)
        else:
            raise ValueError(f"Invalid param: {request.param}")

        return eqx.tree_at(lambda x: x.subjects[s0].admissions[0], tvx_ehr_segmented, a1)

    def test_ehr_serialization(
        self,
        tvx_ehr_concept: rx.TVxEHR,
        tvx_ehr_segmented: rx.TVxEHR,
        mutated_ehr_concept: rx.TVxEHR,
        tmpdir: str,
    ):
        ehr_list = [tvx_ehr_concept, tvx_ehr_segmented, mutated_ehr_concept]
        for i, ehr_i in enumerate(ehr_list):
            path = f"{tmpdir}/tvx_ehr_{i}"
            ehr_i.save(path)
            loaded_ehr_i = rx.TVxEHR.load(path)
            for j in range(i + 1, len(ehr_list)):
                ehr_j = ehr_list[j]
                assert loaded_ehr_i.equals(ehr_j) == (i == j)


class TestObsTimeBinning:
    @pytest.fixture(scope="class")
    def tvx_ehr_concept(self, tvx_ehr: rx.TVxEHR):
        tvx_ehr = eqx.tree_at(lambda x: x.config.interventions, tvx_ehr, False)
        tvx_ehr = eqx.tree_at(lambda x: x.config.observables, tvx_ehr, True)
        return tvx_ehr._execute_pipeline([rx.TVxConcepts()], DATASET_SCHEME_MANAGER)

    @pytest.fixture(scope="class")
    def tvx_ehr_binned(self, tvx_ehr_concept: rx.TVxEHR) -> rx.TVxEHR:
        tvx_ehr_concept = eqx.tree_at(
            lambda x: x.config.time_binning, tvx_ehr_concept, 12.0, is_leaf=lambda x: x is None
        )
        return tvx_ehr_concept._execute_pipeline([rx.ObsTimeBinning()], DATASET_SCHEME_MANAGER)

    def test_binning(self, tvx_ehr_concept: rx.TVxEHR, tvx_ehr_binned: rx.TVxEHR):
        assert all(
            (np.diff(o.time) == tvx_ehr_binned.config.time_binning).all()
            for o in tvx_ehr_binned.iter_obs()
            if len(o) > 1
        )


class TestLeadExtraction:
    @pytest.fixture(scope="class")
    def tvx_ehr_concept(self, tvx_ehr: rx.TVxEHR):
        tvx_ehr = eqx.tree_at(lambda x: x.config.interventions, tvx_ehr, False)
        tvx_ehr = eqx.tree_at(lambda x: x.config.observables, tvx_ehr, True)
        obs_scheme = tvx_ehr.scheme_proxy(DATASET_SCHEME_MANAGER).obs
        lead_config = rx.LeadingObservableExtractorConfig(
            leading_hours=[1.0, 2.0],
            scheme=obs_scheme.name,
            entry_neglect_window=0.0,
            recovery_window=0.0,
            minimum_acquisitions=0,
            observable_code=obs_scheme.index2code[BINARY_OBSERVATION_CODE_INDEX],
        )
        tvx_ehr = eqx.tree_at(lambda x: x.config.leading_observable, tvx_ehr, lead_config, is_leaf=lambda x: x is None)
        return tvx_ehr._execute_pipeline([rx.TVxConcepts()], DATASET_SCHEME_MANAGER)

    @pytest.fixture(scope="class")
    def tvx_ehr_lead(self, tvx_ehr_concept: rx.TVxEHR) -> rx.TVxEHR:
        return tvx_ehr_concept._execute_pipeline([rx.LeadingObservableExtraction()], DATASET_SCHEME_MANAGER)

    def test_lead(self, tvx_ehr_concept: rx.TVxEHR, tvx_ehr_lead: rx.TVxEHR):
        first_patient0 = next(iter(tvx_ehr_concept.subjects.values()))
        first_patient1 = next(iter(tvx_ehr_lead.subjects.values()))
        assert isinstance(first_patient0, rx.Patient)
        assert isinstance(first_patient1, rx.Patient)
        assert isinstance(first_patient0.admissions[0], rx.Admission)
        assert isinstance(first_patient1.admissions[0], rx.Admission)
        assert isinstance(first_patient0.admissions[0].observables, rx.InpatientObservables)
        assert isinstance(first_patient1.admissions[0].observables, rx.InpatientObservables)
        assert first_patient0.admissions[0].leading_observable is None
        assert isinstance(first_patient1.admissions[0].leading_observable, rx.InpatientObservables)

        for subject_id, patient in tvx_ehr_lead.subjects.items():
            for admission in patient.admissions:
                assert admission.leading_observable is not None
                assert admission.observables is not None
                assert len(admission.leading_observable) == len(admission.observables)
                assert admission.leading_observable.value.shape[1] == len(
                    tvx_ehr_lead.config.leading_observable.leading_hours
                )

    def test_ehr_serialization(self, tvx_ehr_concept: rx.TVxEHR, tvx_ehr_lead: rx.TVxEHR, tmpdir: str):
        ehr_list = [tvx_ehr_concept, tvx_ehr_lead]
        for i, ehr_i in enumerate(ehr_list):
            path = f"{tmpdir}/tvx_ehr_{i}"
            ehr_i.save(path)
            loaded_ehr_i = rx.TVxEHR.load(path)
            for j in range(i + 1, len(ehr_list)):
                ehr_j = ehr_list[j]
                assert loaded_ehr_i.equals(ehr_j) == (i == j)

    @pytest.fixture(params=[(lambda tvx: tvx.dataset.tables, rx.DatasetTables), (lambda tvx: tvx.subjects, dict)])
    def tvx_ehr_defer_getters(self, request) -> tuple[Callable[[rx.TVxEHR], Any], type]:
        return request.param

    @pytest.fixture
    def tvx_ehr_lazy_loaded(
        self, tvx_ehr_lead, tmpdir: str, tvx_ehr_defer_getters: tuple[Callable[[rx.TVxEHR], Any], type]
    ) -> rx.TVxEHR:
        getter, _ = tvx_ehr_defer_getters
        tvx_ehr_lead.save(f"{tmpdir}/tvx_ehr_lead.h5", complevel=0)
        return tvx_ehr_lead.load(f"{tmpdir}/tvx_ehr_lead.h5", defer=(getter,), levels=(1,))

    def test_lazy_loading(
        self,
        tvx_ehr_lead: rx.TVxEHR,
        tvx_ehr_lazy_loaded: rx.TVxEHR,
        tvx_ehr_defer_getters: tuple[Callable[[rx.TVxEHR], Any], type],
    ):
        getter, node_type = tvx_ehr_defer_getters
        assert isinstance(getter(tvx_ehr_lazy_loaded), node_type)
        assert all(
            isinstance(child, rx.HDFVirtualNode) for child in eqx.tree_flatten_one_level(getter(tvx_ehr_lazy_loaded))[0]
        )
        loaded = rx.fetch_all(tvx_ehr_lazy_loaded)
        assert loaded.equals(tvx_ehr_lead)

    def test_lazy_loading_subjects(self, tvx_ehr_lead, tmpdir: str):
        tvx_ehr_lead.save(f"{tmpdir}/tvx_ehr_lead.h5", complevel=0)
        lazy_loaded = tvx_ehr_lead.load(f"{tmpdir}/tvx_ehr_lead.h5", defer=(lambda x: x.subjects,), levels=(1,))

        n = len(tvx_ehr_lead.subject_ids)
        splits = tvx_ehr_lead.subject_ids[: n // 2], tvx_ehr_lead.subject_ids[n // 2 :]
        for split in splits:
            lazy_loaded = lazy_loaded.fetch_subjects(split)
            for subject_id in split:
                assert lazy_loaded.subjects[subject_id].equals(tvx_ehr_lead.subjects[subject_id])


#
# class TestWideTVxEHR:
#     BAD_PERFORMANCE_N_ATTRS = 5000
#     BAD_PERFORMANCE_N_CHILDREN = 17000
#
#     @pytest.fixture(scope='class')
#     def wide_tvx_ehr(self, segmented_patient: rx.SegmentedPatient):
#         def array_thinner(x: Any):
#             if isinstance(x, rx.Array):
#                 return np.random.normal(size=())
#             else:
#                 return x
#
#         def obs_thinner(x: rx.InpatientObservables):
#             return rx.InpatientObservables.empty(5)
#
#         segmented_patient = jtu.tree_map(obs_thinner, segmented_patient,
#                                          is_leaf=lambda x: isinstance(x, rx.InpatientObservables))
#         subjects = {f'k_{i}': jtu.tree_map(array_thinner, segmented_patient)
#                     for i in range(self.BAD_PERFORMANCE_N_CHILDREN)}
#         return rx.TVxEHR(dataset=None, config=None, subjects=subjects)
#
#     @pytest.mark.filterwarnings('error:.*maximum number of.*')
#     def test_wide_tvx_ehr_serialization_performance_warnings(self, wide_tvx_ehr: rx.TVxEHR, tmpdir: str):
#         wide_tvx_ehr.save(f'{tmpdir}/wide_tvx_ehr.h5', complevel=3, complib='lzo')
#         loaded = wide_tvx_ehr.load(f'{tmpdir}/wide_tvx_ehr.h5')
#         assert wide_tvx_ehr.equals(loaded)



from dataclasses import fields
from typing import Self

from ..base import AbstractConfig
from ..coding_scheme import CodingSchemesManager, FilterOutcomeMapData
from .ccs import CCSMapRegistration, FlatCCS2ICD9MapOps, MultiLevelCCSICD9MapOps
from .icd import ICDMapOps
from .icd9 import ICD9CMFactory, ICD9PCSFactory
from .icd10 import ICD10CMFactory, ICD10PCSFactory


class Flags(AbstractConfig):
    @classmethod
    def all(cls) -> Self:
        return cls(**{f.name: True for f in fields(cls)})  # type: ignore

    @property
    def flag_set(self) -> tuple[str, ...]:
        return tuple(f.name for f in fields(self) if isinstance(getattr(self, f.name), bool) and getattr(self, f.name))


class ICDSchemeSelection(Flags):
    icd9cm: bool
    icd9pcs: bool
    icd10cm: bool
    icd10pcs: bool

    def __init__(self, icd9cm: bool = False, icd9pcs: bool = False, icd10cm: bool = False, icd10pcs: bool = False):
        self.icd9cm = icd9cm
        self.icd10cm = icd10cm
        self.icd9pcs = icd9pcs
        self.icd10pcs = icd10pcs


class CCSSchemeSelection(Flags):
    dx_ccs: bool
    pr_ccs: bool
    dx_flat_ccs: bool
    pr_flat_ccs: bool

    def __init__(
        self, dx_ccs: bool = False, pr_ccs: bool = False, dx_flat_ccs: bool = False, pr_flat_ccs: bool = False
    ):
        self.dx_ccs = dx_ccs
        self.pr_ccs = pr_ccs
        self.dx_flat_ccs = dx_flat_ccs
        self.pr_flat_ccs = pr_flat_ccs


class OutcomeSelection(Flags):
    icd9cm_v1: bool
    icd9cm_v2_groups: bool
    icd9cm_v3_groups: bool
    dx_flat_ccs_mlhc_groups: bool
    dx_flat_ccs_v1: bool

    def __init__(
        self,
        icd9cm_v1: bool = False,
        icd9cm_v2_groups: bool = False,
        icd9cm_v3_groups: bool = False,
        dx_flat_ccs_mlhc_groups: bool = False,
        dx_flat_ccs_v1: bool = False,
    ):
        self.icd9cm_v1 = icd9cm_v1
        self.icd9cm_v2_groups = icd9cm_v2_groups
        self.icd9cm_v3_groups = icd9cm_v3_groups
        self.dx_flat_ccs_mlhc_groups = dx_flat_ccs_mlhc_groups
        self.dx_flat_ccs_v1 = dx_flat_ccs_v1


def setup_icd_schemes(icd_selection: ICDSchemeSelection) -> CodingSchemesManager:
    manager = CodingSchemesManager()
    if icd_selection.icd9cm:
        manager = manager.add_scheme(ICD9CMFactory.create_scheme())
    if icd_selection.icd9pcs:
        manager = manager.add_scheme(ICD9PCSFactory.create_scheme())
    if icd_selection.icd10cm:
        manager = manager.add_scheme(ICD10CMFactory.create_scheme())
    if icd_selection.icd10pcs:
        manager = manager.add_scheme(ICD10PCSFactory.create_scheme())
    return manager


def setup_ccs_schemes(manager: CodingSchemesManager, ccs_selection: CCSSchemeSelection) -> CodingSchemesManager:
    if ccs_selection.dx_ccs:
        manager = manager.add_scheme(MultiLevelCCSICD9MapOps.create_dx_ccs())
    if ccs_selection.pr_ccs:
        manager = manager.add_scheme(MultiLevelCCSICD9MapOps.create_pr_ccs())
    if ccs_selection.dx_flat_ccs:
        manager = manager.add_scheme(FlatCCS2ICD9MapOps.create_dx_flat_ccs())
    if ccs_selection.pr_flat_ccs:
        manager = manager.add_scheme(FlatCCS2ICD9MapOps.create_pr_flat_ccs())
    return manager


def setup_outcomes(manager: CodingSchemesManager, outcome_selection: OutcomeSelection) -> CodingSchemesManager:
    for outcome_name in outcome_selection.flag_set:
        manager = manager.add_outcome(FilterOutcomeMapData.from_spec_json(manager.scheme, f"{outcome_name}.json"))
    return manager


def setup_icd_icd_maps(manager: CodingSchemesManager, scheme_selection: ICDSchemeSelection) -> CodingSchemesManager:
    # ICD9 <-> ICD10s
    if scheme_selection.icd9cm and scheme_selection.icd10cm:
        manager = ICDMapOps.register_mappings(manager, "icd10cm", "icd9cm", "2018_gem_cm_I10I9.txt.gz")
        manager = ICDMapOps.register_mappings(manager, "icd9cm", "icd10cm", "2018_gem_cm_I9I10.txt.gz")

    if scheme_selection.icd10pcs and scheme_selection.icd9pcs:
        manager = ICDMapOps.register_mappings(manager, "icd10pcs", "icd9pcs", "2018_gem_pcs_I10I9.txt.gz")
        manager = ICDMapOps.register_mappings(manager, "icd9pcs", "icd10pcs", "2018_gem_pcs_I9I10.txt.gz")
    return manager


def setup_icd_ccs_maps(
    manager: CodingSchemesManager, icd_selection: ICDSchemeSelection, ccs_selection: CCSSchemeSelection
) -> CodingSchemesManager:
    """
    Enables the following two-way mappings (if all selected):

    |                   | icd9cm   | icd9pcs   | icd10cm  | dx_flat_icd10 | icd9pcs |
    |-------------------|-----------|-----------|-----------|---------------|---------------|
    | dx_ccs            |   X       |           |   X       |      X        |               |
    | dx_flat_ccs       |   X       |           |   X       |      X        |               |
    | pr_ccs            |           |   X       |           |               |      X        |
    | pr_flat_ccs       |           |   X       |           |               |      X        |

    In addition to dx_ccs <-> dx_flat_ccs and pr_ccs <-> pr_flat_ccs.
    """

    if icd_selection.icd9cm:
        manager = CCSMapRegistration.icd9cm_maps(
            manager, "icd9cm", dx_ccs=ccs_selection.dx_ccs, dx_flat_ccs=ccs_selection.dx_flat_ccs
        )
    if icd_selection.icd9pcs:
        manager = CCSMapRegistration.icd9pcs_maps(
            manager, "icd9pcs", pr_ccs=ccs_selection.pr_ccs, pr_flat_ccs=ccs_selection.pr_flat_ccs
        )
    if icd_selection.icd10cm:
        manager = CCSMapRegistration.icd10cm_maps(
            manager, "icd10cm", dx_ccs=ccs_selection.dx_ccs, dx_flat_ccs=ccs_selection.dx_flat_ccs
        )
    if icd_selection.icd10pcs:
        manager = CCSMapRegistration.icd10pcs_maps(
            manager, "icd10pcs", pr_ccs=ccs_selection.pr_ccs, pr_flat_ccs=ccs_selection.pr_flat_ccs
        )

    # cross-maps
    dx_bridge, pr_bridge = None, None
    if icd_selection.icd9cm and ccs_selection.dx_ccs and ccs_selection.dx_flat_ccs:
        dx_bridge = "icd9cm"
    if icd_selection.icd9pcs and ccs_selection.pr_ccs and ccs_selection.pr_flat_ccs:
        pr_bridge = "icd9pcs"
    manager = CCSMapRegistration.ccs_flat_to_multi_maps(manager, dx_bridge=dx_bridge, pr_bridge=pr_bridge)

    return manager


def setup_standard_icd_ccs(
    icd_selection: ICDSchemeSelection = ICDSchemeSelection.all(),
    ccs_selection: CCSSchemeSelection = CCSSchemeSelection.all(),
    outcome_selection: OutcomeSelection = OutcomeSelection.all(),
) -> CodingSchemesManager:
    manager = setup_icd_schemes(icd_selection)
    manager = setup_icd_icd_maps(manager, icd_selection)
    manager = setup_ccs_schemes(manager, ccs_selection)
    manager = setup_icd_ccs_maps(manager, icd_selection, ccs_selection)
    manager = setup_outcomes(manager, outcome_selection)
    return manager



from .base import (  # noqa
    AbstractConfig,
    AbstractModule,
    AbstractVxData,
    AbstractWithDataframeEquivalent,
    AbstractWithSeriesEquivalent,
    HDFVirtualNode,
    fetch_all,
    fetch_at,
    fetch_one_level_at,
)

from .coding_scheme import (  # noqa
    CodeMap,
    CodesVector,
    CodingScheme,
    CodingSchemeWithUOM,
    HierarchicalScheme,
    CodingSchemesManager,
    NumericScheme,
    NumericalTypeHint,
    FilterOutcomeMapData,
    FilterOutcomeMap,
    ReducedCodeMapN1,
    GroupingData,
    AggregationLiteral,
)

from .dataset import (  # noqa
    COLUMN,
    Dataset,
    DatasetConfig,
    DatasetSchemeConfig,
    DatasetSchemeProxy,
    DatasetTables,
    DatasetColumns,
    Report,
    ReportAttributes,
    SplitLiteral,
    AbstractDatasetPipeline,
)

from .freezer import (  # noqa
    FrozenDict11,
    FrozenDict1N,
    FrozenDict1NM,
)

from .transformations import (  # noqa
    CastTimestamps,
    DatasetTransformation,
    FilterClampTimestampsToAdmissionInterval,
    FilterInvalidInputRatesSubjects,
    FilterSubjectsNegativeAdmissionLengths,
    FilterSubjectsWithInvalidInputInterval,
    FilterUnsupportedCodes,
    ICUInputRateUnitConversion,
    MergeOverlappingAdmissions,
    RemoveSubjectsWithOverlappingAdmissions,
    SetAdmissionRelativeTimes,
    SetIndex,
    SynchronizeSubjects,
    FilterSubjectsWithLongAdmission,
    FilterAdmissionsWithNoDiagnoses,
    FilterAdmissionsWithNoObservables,
    FilterSubjectsWithSingleOrNoAdmission,
    FilterShortAdmissions,
    SqueezeToStandardColumns,
)

from .tvx_concepts import (  # noqa
    Admission,
    AdmissionDates,
    DemographicVectorConfig,
    InpatientInput,
    InpatientInterventions,
    InpatientObservables,
    LeadingObservableExtractor,
    LeadingObservableExtractorConfig,
    Patient,
    SegmentedAdmission,
    SegmentedInpatientInterventions,
    SegmentedInpatientObservables,
    SegmentedPatient,
    StaticInfo,
)

from .tvx_ehr import (  # noqa
    DatasetNumericalProcessors,
    DatasetNumericalProcessorsConfig,
    IQROutlierRemoverConfig,
    OutlierRemoversConfig,
    ScalerConfig,
    ScalersConfig,
    SegmentedTVxEHR,
    TVxEHR,
    TVxEHRConfig,
    TVxEHRSampleConfig,
    TVxEHRSchemeConfig,
    TVxReport,
)

from .tvx_transformations import (  # noqa
    CodedValueScaler,
    InputScaler,
    InterventionSegmentation,
    LeadingObservableExtraction,
    ObsAdaptiveScaler,
    ObsIQROutlierRemover,
    ObsTimeBinning,
    SampleSubjects,
    TrainableTransformation,
    TVxConcepts,
)

from .utils import (  # noqa
    Array,
    ArrayTypes,
    path_from_getter,
    path_from_jax_keypath,
    load_config,
    write_config,
    translate_path,
)



import ehrax as rx
import pytest
from jax._src.tree_util import DictKey, GetAttrKey, SequenceKey


def test_path_from_jax():
    kpath = (SequenceKey(4), DictKey(5), DictKey("6"), DictKey("x"), GetAttrKey("as"))
    assert rx.path_from_jax_keypath(kpath) == ["4", "5", "6", "x", "as"]


@pytest.mark.parametrize(
    "lambda_, expected",
    [
        (lambda x: x, []),
        (lambda x: x.y, ["y"]),
        (lambda x: x[100], ["100"]),
        (lambda x: x.s.m.n[34].at["d"], ["s", "m", "n", "34", "at", "d"]),
    ],
)
def test_path_from_getter(lambda_, expected):
    assert rx.path_from_getter(lambda_) == expected



# """."""
#
# import logging
# from collections import defaultdict
# from concurrent.futures import ThreadPoolExecutor
#
# import pandas as pd
#
# from ._dataset_mimic3 import MIMIC3Dataset
# from .coding_scheme import CodingScheme
# from .concepts import (Patient, Admission, CPRDDemographicVectorConfig, CPRDStaticInfo)
# from .dataset import DatasetScheme, DatasetSchemeConfig
# from ..utils import translate_path
#
#
# class CPRDDatasetSchemeConfig(DatasetSchemeConfig):
#     imd: str = "CPRDIMDCategorical"
#
#
# class CPRDDatasetScheme(DatasetScheme):
#     imd: CodingScheme
#
#     def __init__(self, config: CPRDDatasetSchemeConfig, **kwargs):
#         super().__init__(config, **kwargs)
#         self.imd = CodingScheme.from_name(config.imd)
#
#     def demographic_vector_size(
#             self, demographic_vector_config: CPRDDemographicVectorConfig):
#         size = DatasetScheme.demographic_vector_size(
#             self, demographic_vector_config)
#         if demographic_vector_config.imd:
#             size += len(self.imd)
#         return size
#
#
# class CPRDDataset(MIMIC3Dataset):
#
#     def subject_info_extractor(self, subject_ids, target_scheme):
#
#         static_df = self.df['static']
#         c_gender = self.colname["static"].gender
#         c_eth = self.colname["static"].ethnicity
#         c_imd = self.colname["static"].imd_decile
#         c_dob = self.colname["static"].date_of_birth
#
#         static_df = static_df.loc[subject_ids]
#         gender = static_df[c_gender].map(self.scheme.gender.codeset2vec)
#         subject_gender = gender.to_dict()
#
#         subject_dob = static_df[c_dob].dt.normalize().to_dict()
#         subject_eth = dict()
#         eth_mapper = self.scheme.ethnicity_mapper(target_scheme)
#         subject_eth = static_df[c_eth].map(lambda eth: eth_mapper.codeset2vec(
#             eth_mapper.map_codeset(eth))).to_dict()
#
#         subject_imd = static_df[c_imd].map(
#             self.scheme.imd.codeset2vec).to_dict()
#         return subject_dob, subject_gender, subject_eth, subject_imd
#
#     def _load_dataframes(self):
#         config = self.config.copy()
#         colname = self.colname["adm"]
#
#         df = pd.read_csv(translate_path(config.path), sep='\t', dtype=str)
#
#         def listify(s):
#             return list(map(lambda e: e.strip(), s.split(',')))
#
#         adm_tups = []
#         dx_tups = []
#         demo_tups = []
#         admission_id = 0
#         for subject_id, _subj_df in df.groupby(colname.subject_id):
#             subject_id = int(subject_id)
#             assert len(_subj_df) == 1, "Each patient should have a single row"
#             subject = _subj_df.iloc[0].to_dict()
#             codes = listify(subject[colname.code])
#             year_month = listify(subject[colname.dischtime])
#
#             # To infer date-of-birth
#             age0 = int(float(listify(subject[colname.age_at_dischtime])[0]))
#             year_month0 = pd.to_datetime(year_month[0]).normalize()
#             date_of_birth = year_month0 + pd.DateOffset(years=-age0)
#             gender = subject[colname.gender]
#             imd = subject[colname.imd_decile]
#             ethnicity = subject[colname.ethnicity]
#             demo_tups.append(
#                 (subject_id, date_of_birth, gender, imd, ethnicity))
#             # codes aggregated by year-month.
#             dx_codes_ym_agg = defaultdict(set)
#
#             for code, ym in zip(codes, year_month):
#                 ym = pd.to_datetime(ym).normalize()
#                 dx_codes_ym_agg[ym].add(code)
#             for disch_date in sorted(dx_codes_ym_agg.keys()):
#                 admit_date = disch_date + pd.DateOffset(days=-1)
#                 adm_tups.append(
#                     (subject_id, admission_id, admit_date, disch_date))
#
#                 dx_codes = dx_codes_ym_agg[disch_date]
#                 dx_tups.extend([(admission_id, dx_code)
#                                 for dx_code in dx_codes])
#                 admission_id += 1
#
#         adm_keys = ('subject_id', 'admission_id', 'admittime', 'dischtime')
#         dx_keys = ('admission_id', 'code')
#         demo_keys = ('subject_id', 'date_of_birth', 'gender', 'imd_decile',
#                      'ethnicity')
#
#         adm_cols = ColumnNames.make({k: colname.get(k, k) for k in adm_keys})
#         dx_cols = ColumnNames.make({k: colname.get(k, k) for k in dx_keys})
#         demo_cols = ColumnNames.make({k: colname.get(k, k) for k in demo_keys})
#         adm_cols = adm_cols._replace(index=adm_cols.admission_id)
#         demo_cols = demo_cols._replace(index=demo_cols.subject_id)
#
#         adm_df = pd.DataFrame(adm_tups,
#                               columns=list(map(adm_cols.get, adm_keys)))
#         adm_df = adm_df.astype({
#             adm_cols.admission_id: int,
#             adm_cols.subject_id: int
#         }).set_index(adm_cols.index)
#
#         dx_df = pd.DataFrame(dx_tups, columns=list(map(dx_cols.get, dx_keys)))
#         dx_df = dx_df.astype({dx_cols.admission_id: int})
#
#         demo_df = pd.DataFrame(demo_tups,
#                                columns=list(map(demo_cols.get, demo_keys)))
#         demo_df = demo_df.astype({
#             demo_cols.subject_id: int
#         }).set_index(demo_cols.index)
#
#         self.df = {'adm': adm_df, 'dx_discharge': dx_df, 'static': demo_df}
#         self.colname = {'adm': adm_cols, 'dx_discharge': dx_cols, 'static': demo_cols}
#         self._match_admissions_with_demographics(self.df, self.colname)
#
#     def to_subjects(self, subject_ids, num_workers, demographic_vector_config,
#                     target_scheme: DatasetScheme, **kwargs):
#
#         (subject_dob, subject_gender, subject_eth,
#          subject_imd) = self.subject_info_extractor(subject_ids, target_scheme)
#         admission_ids = self.adm_extractor(subject_ids)
#         adm_ids_list = sum(map(list, admission_ids.values()), [])
#         logging.debug('Extracting dx_discharge codes...')
#         dx_codes = dict(self.dx_codes_extractor(adm_ids_list, target_scheme))
#         logging.debug('[DONE] Extracting dx_discharge codes')
#         logging.debug('Extracting dx_discharge codes history...')
#         dx_codes_history = dict(
#             self.dx_codes_history_extractor(dx_codes, admission_ids,
#                                             target_scheme))
#         logging.debug('[DONE] Extracting dx_discharge codes history')
#         logging.debug('Extracting outcome...')
#         outcome = dict(self.outcome_extractor(dx_codes, target_scheme))
#         logging.debug('[DONE] Extracting outcome')
#
#         logging.debug('Compiling admissions...')
#         c_admittime = self.colname['adm'].admittime
#         c_dischtime = self.colname['adm'].dischtime
#         adf = self.df['adm']
#         adm_dates = dict(
#             zip(adf.index, zip(adf[c_admittime], adf[c_dischtime])))
#
#         def gen_admission(i):
#             return Admission(admission_id=i,
#                              admission_dates=adm_dates[i],
#                              dx_codes=dx_codes[i],
#                              dx_codes_history=dx_codes_history[i],
#                              outcome=outcome[i],
#                              observables=None,
#                              interventions=None)
#
#         def _gen_subject(subject_id):
#             _admission_ids = admission_ids[subject_id]
#             # for subject_id, subject_admission_ids in admission_ids.items():
#             _admission_ids = sorted(_admission_ids,
#                                     key=lambda aid: adm_dates[aid][0])
#             static_info = CPRDStaticInfo(
#                 date_of_birth=subject_dob[subject_id],
#                 gender=subject_gender[subject_id],
#                 ethnicity=subject_eth[subject_id],
#                 imd=subject_imd[subject_id],
#                 demographic_vector_config=demographic_vector_config)
#
#             with ThreadPoolExecutor(max_workers=num_workers) as executor:
#                 admissions = list(executor.map(gen_admission, _admission_ids))
#             return Patient(subject_id=subject_id,
#                            admissions=admissions,
#                            static_info=static_info)
#
#         return list(map(_gen_subject, subject_ids))



import logging

import pandas as pd

from ..coding_scheme import CodeMap, CodingScheme, CodingSchemesManager, Formatter, FrozenDict1N, HierarchicalScheme
from ..utils import dataframe_log, resources_path


class ICDScheme(CodingScheme, Formatter):
    # def __check_init__(self):
    #     # assert formatted codes.
    #     assert tuple(map(self.reformat, self.legacy_map.keys())) == tuple(self.legacy_map.keys())
    #     assert tuple(map(self.reformat, self.legacy_map.values())) == tuple(self.legacy_map.values())
    #     # assert existence of values in codes.
    #     assert all(c in self.codes for c in self.legacy_map.values())

    @staticmethod
    def deformat(code: str) -> str:
        return code.strip().replace(".", "")


class ICDHierarchicalScheme(HierarchicalScheme, ICDScheme):
    pass


class ICDMapOps:
    @staticmethod
    def load_conversion_table(conversion_filename: str) -> pd.DataFrame:
        df = pd.read_csv(
            resources_path("ICD", conversion_filename), sep=r"\s+", dtype=str, names=["source", "target", "meta"]
        )
        df["approximate"] = df["meta"].apply(lambda s: s[0])
        df["no_map"] = df["meta"].apply(lambda s: s[1])
        df["combination"] = df["meta"].apply(lambda s: s[2])
        df["scenario"] = df["meta"].apply(lambda s: s[3])
        df["choice_list"] = df["meta"].apply(lambda s: s[4])
        return df

    @staticmethod
    def conversion_status(conversion_table: pd.DataFrame) -> dict[str, str]:
        def _get_status(groupby_df: pd.DataFrame):
            if (groupby_df["no_map"] == "1").all():
                return "no_map"
            elif len(groupby_df) == 1:
                return "11_map"
            elif groupby_df["scenario"].nunique() > 1:
                return "ambiguous"
            elif groupby_df["choice_list"].nunique() < len(groupby_df):
                return "1n_map(resolved)"
            else:
                return "1n_map"

        return conversion_table.groupby("source")[["no_map", "scenario", "choice_list"]].apply(_get_status).to_dict()

    @staticmethod
    def register_mappings(
        manager: CodingSchemesManager, source_scheme: str, target_scheme: str, conversion_filename: str
    ) -> CodingSchemesManager:  # expose
        logging.debug(f"[BEGIN] mapping from {source_scheme} to {target_scheme} via {conversion_filename}.")
        source_scheme = manager.scheme[source_scheme]
        target_scheme = manager.scheme[target_scheme]
        assert isinstance(source_scheme, ICDScheme) and isinstance(target_scheme, ICDScheme), (
            f"Expected ICDScheme subclasses. Got {type(source_scheme)} and {type(target_scheme)} instead."
        )
        df = ICDMapOps.load_conversion_table(conversion_filename=conversion_filename).reset_index(drop=True)
        df = df.assign(source=df["source"].map(source_scheme.reformat), target=df["target"].map(target_scheme.reformat))
        valid_target = df["target"].isin(target_scheme.index).values
        valid_source = df["source"].isin(source_scheme.index).values

        table = df[valid_target & valid_source]
        conversion_status = ICDMapOps.conversion_status(table)
        table = table.assign(status=table["source"].map(conversion_status))
        table = table[table["status"] != "no_map"]
        data = FrozenDict1N(table.groupby("source")["target"].apply(set).to_dict())
        m = CodeMap(source_name=source_scheme.name, target_name=target_scheme.name, data=data)
        # Wait. Let's report and log.
        report = df[(~valid_target) | (~valid_source)]
        report = report.assign(
            invalid_target=~valid_target[report.index.values], invalid_source=~valid_source[report.index.values]
        )
        dataframe_log.info(
            f"In processing {conversion_filename}. "
            f"{len(m.domain)} (of {len(source_scheme)}) {m.source_name} source codes were mapped to "
            f"{len(m.range)} {m.target_name} (of {len(target_scheme)}) target codes. "
            f"{(~valid_source).sum()} (of {len(valid_source)}) source code in the conversion table were discarded. "
            f"{(~valid_target).sum()} (of {len(valid_target)}) target code in the conversion table were discarded. ",
            dataframe=report,
            tag=f"conversion_miss_report_{source_scheme.name}_{target_scheme.name}",
        )
        logging.debug(f"[DONE] mapping from {source_scheme} to {target_scheme} via {conversion_filename}.")
        # Carry on. Done report and log.
        return manager.add_map(m)



from collections.abc import Mapping
from dataclasses import dataclass
from typing import get_args, TYPE_CHECKING

import numpy as np
import pandas as pd
import scipy.stats.distributions as dist

from .._literals import TableAggregationLiteral
from ..coding_scheme import CodeMap, CodingSchemesManager


if TYPE_CHECKING:
    from ..dataset import AdmissionsTableColumns, Dataset, DatasetSchemeProxy  # type: ignore


@dataclass
class TargetHistogram:
    dataset: "Dataset"
    schemes_manager: CodingSchemesManager

    def __init__(self, dataset: "Dataset", schemes_manager: CodingSchemesManager):
        self.dataset = dataset
        self.schemes_manager = schemes_manager

    @staticmethod
    def compute(table: pd.DataFrame, c_admission_id: str, c_code: str, scheme_mapper: CodeMap) -> pd.Series:
        s_table = table[[c_admission_id, c_code]].drop_duplicates()
        s_table = s_table[s_table[c_code].isin(scheme_mapper.domain)]
        # group by admission_id to collapse duplicate target codes within a single admission.
        t_table = s_table.groupby(c_admission_id)[c_code].apply(scheme_mapper.map_codeset)
        return t_table.explode().value_counts()

    @staticmethod
    def adapt_aggregation_level(
        admissions: pd.DataFrame,
        admissions_cols: "AdmissionsTableColumns",
        table: pd.DataFrame,
        c_admission_id: str,
        aggregation_level: TableAggregationLiteral,
    ) -> tuple[pd.DataFrame, int]:
        c_subject_id = admissions_cols.subject_id
        match aggregation_level:
            case "admission":
                # do nothing.
                return table, len(admissions)
            case "first_admission":
                # Apply the statistics only on the first admission for each subject.
                # Collect the first admission id for each subject and remove the rest.
                admissions = admissions.sort_values(by=admissions_cols.start_time, ascending=True)
                admission_index = str(admissions.index.name)
                assert admission_index == c_admission_id
                first_admissions = admissions.reset_index(drop=False).groupby(c_subject_id)[admission_index].first()
                return table[table[c_admission_id].isin(first_admissions)], len(first_admissions)
            case "subject":
                # Apply the statistics on the level of each subject as a whole.
                # To adapt to the same function of `compute`, we just rename admission ids of each subject
                # to have the same dummy value. We just set the values of admission ids to the subject ids.
                table = table.assign(**{c_admission_id: table[c_admission_id].map(admissions[c_subject_id].to_dict())})
                return table, admissions[c_subject_id].nunique()
            case _:
                raise ValueError(
                    f"Unknown aggregation level '{aggregation_level}'. "
                    f"Expected one of: {get_args(TableAggregationLiteral)}."
                )

    def _dx_discharge(
        self, codemap: CodeMap, target_codes: tuple[str, ...], aggregation_level: TableAggregationLiteral = "admission"
    ) -> tuple[pd.Series, int]:
        table = self.dataset.tables.dx_discharge
        cols = self.dataset.config.columns.dx_discharge
        table, n = self.adapt_aggregation_level(
            self.dataset.tables.admissions,
            self.dataset.config.columns.admissions,
            table,
            cols.admission_id,
            aggregation_level,
        )
        hist = self.compute(table, cols.admission_id, cols.code, codemap)
        return hist.reindex(pd.Index(target_codes, name=cols.code), fill_value=0), n

    def dx_discharge(
        self, target_scheme: str | tuple[str, ...], aggregation_level: TableAggregationLiteral = "admission"
    ) -> tuple[pd.Series, int]:
        if isinstance(target_scheme, str):
            codemap = self.schemes_manager.map[self.dataset.config.scheme.dx_discharge, target_scheme]
        elif isinstance(target_scheme, tuple):
            path = (self.dataset.config.scheme.dx_discharge,) + target_scheme
            codemap = self.schemes_manager.make_chained_map(path)
            target_scheme = target_scheme[-1]
        else:
            raise ValueError(f"Expected a string or a tuple of strings, got {type(target_scheme)}.")

        codes = self.schemes_manager.scheme[target_scheme].codes
        return self._dx_discharge(codemap, codes, aggregation_level)

    def outcome(
        self, outcome: str | tuple[str, ...], aggregation_level: TableAggregationLiteral = "admission"
    ) -> tuple[pd.Series, int]:
        dx_scheme = self.dataset.config.scheme.dx_discharge
        if isinstance(outcome, str):
            o = self.schemes_manager.outcome[dx_scheme, outcome]
        elif isinstance(outcome, tuple):
            chain, outcome = outcome[:-1], outcome[-1]
            chain = (dx_scheme, *chain, self.schemes_manager.outcome_data[outcome].base_name)
            m = self.schemes_manager.make_chained_map(chain)
            o = self.schemes_manager.add_map(m, overwrite=True).outcome[dx_scheme, outcome]
        else:
            raise ValueError(f"Expected a string or a tuple of strings, got {type(outcome)}.")
        return self._dx_discharge(o.codemap, o.scheme.codes, aggregation_level)


@dataclass
class MultiDatasetTargetHistogram:
    target_hist: tuple[TargetHistogram, ...]
    schemes_manager: CodingSchemesManager

    def __init__(self, datasets: tuple["Dataset", ...], schemes_manager: CodingSchemesManager):
        self.target_hist = tuple(TargetHistogram(dataset, schemes_manager) for dataset in datasets)
        self.schemes_manager = schemes_manager

    @staticmethod
    def compile(results: tuple[pd.Series, ...], normalize: tuple[int, ...]) -> tuple[pd.DataFrame, pd.Series]:
        df = pd.concat(results, axis=1)
        df.columns = [f"D{i}" for i in range(len(results))]
        df.index.name = results[0].index.name
        return df, pd.Series(normalize, index=df.columns)

    def dx_discharge(
        self, target_scheme: str, aggregation_level: TableAggregationLiteral = "admission"
    ) -> tuple[pd.DataFrame, pd.Series]:
        results, norm = zip(*[h.dx_discharge(target_scheme, aggregation_level) for h in self.target_hist])
        return self.compile(results, norm)

    def outcome(
        self, target_scheme: str, aggregation_level: TableAggregationLiteral = "admission"
    ) -> tuple[pd.DataFrame, pd.Series]:
        results, norm = zip(*[h.outcome(target_scheme, aggregation_level) for h in self.target_hist])
        return self.compile(results, norm)


@dataclass
class DatasetStatsInterface:
    dataset: "Dataset"
    schemes_manager: CodingSchemesManager

    def __init__(self, dataset: "Dataset", schemes_manager: CodingSchemesManager):
        self.dataset = dataset
        self.schemes_manager = schemes_manager

    def age_at_first_admission(self) -> pd.Series:
        c_subject_id = self.dataset.config.columns.admissions.subject_id
        c_admittime = self.dataset.config.columns.admissions.start_time
        c_dob = self.dataset.config.columns.static.date_of_birth
        admissions = self.dataset.tables.admissions.sort_values(by=c_admittime, ascending=True)
        first_admission = admissions.groupby(c_subject_id, as_index=False)[[c_subject_id, c_admittime]].first()
        admission_year = first_admission[c_admittime].dt.year.values
        birth_year = self.dataset.tables.static.loc[first_admission[c_subject_id], c_dob].dt.year.values
        return pd.Series(admission_year - birth_year, index=first_admission[c_subject_id])

    @property
    def schemes_proxy(self) -> "DatasetSchemeProxy":
        return self.dataset.scheme_proxy(self.schemes_manager)

    @property
    def target_hist(self) -> TargetHistogram:
        return TargetHistogram(self.dataset, self.schemes_manager)


@dataclass
class TwoDatasetsTargetHistogram(MultiDatasetTargetHistogram):
    def __init__(self, datasets: tuple["Dataset", ...], schemes_manager: CodingSchemesManager):
        super().__init__(datasets, schemes_manager)
        assert len(datasets) == 2


@dataclass
class MultiDatasetsStatsInterface:
    datasets: tuple["Dataset", ...]
    schemes_manager: CodingSchemesManager

    def __init__(self, *datasets: "Dataset", schemes_manager: CodingSchemesManager):
        self.datasets = datasets
        self.schemes_manager = schemes_manager

    @property
    def target_hist(self) -> MultiDatasetTargetHistogram:
        return MultiDatasetTargetHistogram(self.datasets, self.schemes_manager)


@dataclass
class TwoSamplesTest:
    target_hist: TwoDatasetsTargetHistogram

    def __init__(self, target_histogram: TwoDatasetsTargetHistogram):
        self.target_hist = target_histogram

    @staticmethod
    def proportion_tests(counts: pd.DataFrame, n: pd.Series, code_description: Mapping[str, str]) -> pd.DataFrame:
        assert counts.shape[1] == n.shape[0] == 2
        # z-test two sample proportion: https://mverbakel.github.io/2021-02-13/two-sample-proportions
        x1, x2 = counts.iloc[:, 0], counts.iloc[:, 1]
        n1, n2 = n.loc[counts.columns].tolist()
        avg_p = (x1 + x2) / n.sum()
        z_val = (x1 / n1 - x2 / n2) / np.sqrt(avg_p * (1 - avg_p) * (1 / n1 + 1 / n2))
        z_prob = pd.Series(-np.abs(z_val)).map(dist.norm.cdf)
        normal_assumption = (x1 >= 10) & (x2 >= 10) & (n1 - x1 >= 10) & (n2 - x2 >= 10)  # same blog post.
        return pd.DataFrame(
            {
                "code_description": list(map(code_description.get, counts.index)),
                "p_val_two_sided": 2 * z_prob,
                "normal_assumption": normal_assumption,
                "D0_counts": x1,
                "D1_counts": x2,
                "D0_p": x1 / n1,
                "D1_p": x2 / n2,
            },
            index=counts.index,
        )

    @staticmethod
    def summerise(stats: pd.DataFrame, alpha: float = 0.05) -> pd.Series:
        divergent = (stats["normal_assumption"]) & (stats["p_val_two_sided"] < alpha)
        convergent = (stats["normal_assumption"]) & (stats["p_val_two_sided"] > alpha)
        n_divergent = divergent.sum()
        n_convergent = convergent.sum()
        n_skip_test = (~stats["normal_assumption"]).sum()
        n = stats.shape[0]
        return pd.Series({"total": n, "divergent": n_divergent, "convergent": n_convergent, "skip_test": n_skip_test})

    def dx_discharge(
        self, target_scheme: str | tuple[str, ...], aggregation_level: TableAggregationLiteral = "admission"
    ) -> pd.DataFrame:
        counts, n = self.target_hist.dx_discharge(target_scheme, aggregation_level)
        if isinstance(target_scheme, tuple):
            target_scheme = target_scheme[-1]
        return self.proportion_tests(counts, n, self.target_hist.schemes_manager.scheme[target_scheme].desc)

    def outcome(
        self, target_scheme: str | tuple[str, ...], aggregation_level: TableAggregationLiteral = "admission"
    ) -> pd.DataFrame:
        counts, n = self.target_hist.outcome(target_scheme, aggregation_level)
        if isinstance(target_scheme, tuple):
            target_scheme = target_scheme[-1]
        return self.proportion_tests(counts, n, self.target_hist.schemes_manager.outcome_scheme[target_scheme].desc)


@dataclass
class TwoDatasetsStatsInterface:
    datasets: tuple["Dataset", ...]
    schemes_manager: CodingSchemesManager

    def __init__(self, *datasets: "Dataset", schemes_manager: CodingSchemesManager):
        self.datasets = datasets
        self.schemes_manager = schemes_manager
        assert len(datasets) == 2

    @property
    def target_hist(self) -> TwoDatasetsTargetHistogram:
        return TwoDatasetsTargetHistogram(self.datasets, self.schemes_manager)

    @property
    def target_p_tests(self) -> TwoSamplesTest:
        return TwoSamplesTest(self.target_hist)

    @staticmethod
    def summerise_p_tests(stats: pd.DataFrame, alpha: float = 0.05) -> pd.Series:
        return TwoSamplesTest.summerise(stats, alpha)



from typing import Any, Final

import pandas as pd

from ..coding_scheme import FrozenDict11, HierarchicalScheme
from ..utils import resources_path
from .icd import ICDHierarchicalScheme


class ICD9CM(ICDHierarchicalScheme):
    @staticmethod
    def format(code: str) -> str:
        if "-" in code:
            return "-".join(map(ICD9CM.format, code.split("-")))
        if "." in code:
            # logging.debug(f'Code {code} already is in decimal format')
            return code
        n = len(code)
        if code[0] == "E":
            if n > 4:
                return code[:4] + "." + code[4:]
            else:
                return code
        if n > 3:
            return code[:3] + "." + code[3:]
        else:
            return code


class ICD9PCS(ICDHierarchicalScheme):
    @staticmethod
    def format(code: str) -> str:
        if "-" in code:
            return "-".join(map(ICD9PCS.format, code.split("-")))
        if "." in code:
            # logging.debug(f'Code {code} already is in decimal format')
            return code
        if len(code) > 2:
            return code[:2] + "." + code[2:]
        else:
            return code


class ICD9PCSFactory:
    ICD9_FILE: str = resources_path("ICD", "HOM-ICD9.csv.gz")
    DUMMY_ROOT_CLASS_ID: str = "owl#Thing"
    PCS_ROOT_CLASS_ID: str = "MM_CLASS_2"
    CM_ROOT_CLASS_ID: str = "MM_CLASS_21"

    @classmethod
    def create_scheme_data(
        cls,
        processed_icd_table: pd.DataFrame,
        all_parent_to_children_map: dict[str, frozenset[str]],
        select_tree: str | None = None,
        deselect_tree: str | None = None,
    ) -> dict[str, Any]:
        assert (select_tree is None) != (deselect_tree is None), (
            "Must specify exactly one of select_tree and deselect_tree."
        )
        if select_tree is not None:
            # Select a subset. Select the procedure tree if you want a procedure scheme, to discard everything else.
            pt2ch = cls.select_subtree(all_parent_to_children_map, select_tree)  # cls.PR_ROOT_CLASS_ID)
        else:
            # Remove a subset. Remove the procedure tree if you want the diagnostic scheme.
            pt2ch = cls.deselect_subtree(all_parent_to_children_map, deselect_tree)

        # Combining remaining node indices in one set.
        nodes = set(list(pt2ch.keys()) + [n for ns in pt2ch.values() for n in ns])

        # Filter out the procedure code from the df.
        df = processed_icd_table[processed_icd_table["NODE_IDX"].isin(nodes)]
        return cls.generate_dictionaries(df) | {"ch2pt": HierarchicalScheme.reverse_connection(pt2ch)}

    @classmethod
    def create_cm_scheme_data(
        cls, processed_icd_table: pd.DataFrame, all_parent_to_children_map: dict[str, frozenset[str]]
    ) -> dict[str, Any]:
        return cls.create_scheme_data(
            processed_icd_table, all_parent_to_children_map, deselect_tree=cls.PCS_ROOT_CLASS_ID
        )

    @classmethod
    def create_pcs_scheme_data(
        cls, processed_icd_table: pd.DataFrame, all_parent_to_children_map: dict[str, frozenset[str]]
    ) -> dict[str, Any]:
        return cls.create_scheme_data(
            processed_icd_table, all_parent_to_children_map, select_tree=cls.PCS_ROOT_CLASS_ID
        )

    @staticmethod
    def deselect_subtree(pt2ch: dict[str, frozenset[str]], sub_root: str) -> dict[str, frozenset[str]]:
        to_del = HierarchicalScheme._bfs_traversal(pt2ch, sub_root, True)
        to_use = set(pt2ch.keys()) - set(to_del)
        return {pt: pt2ch[pt] for pt in to_use}

    @staticmethod
    def select_subtree(pt2ch: dict[str, frozenset[str]], sub_root: str) -> dict[str, frozenset[str]]:
        to_use = HierarchicalScheme._bfs_traversal(pt2ch, sub_root, True)
        to_use = set(to_use) & set(pt2ch.keys())
        return {pt: pt2ch[pt] for pt in to_use}

    @classmethod
    def load_raw_table(cls) -> pd.DataFrame:
        # https://bioportal.bioontology.org/ontologies/HOM-ICD9
        return pd.read_csv(cls.ICD9_FILE, dtype=str)

    @classmethod
    def extract_icd9_codes(cls, df: pd.DataFrame) -> list[str]:
        return list(df["C_BASECODE"].apply(lambda c: c.split(":")[-1]))

    @classmethod
    def process_icd_table(cls, table: pd.DataFrame) -> pd.DataFrame:
        df = table.fillna("")

        def retain_suffix(cell):
            if "http" in cell:
                return cell.split("/")[-1]
            else:
                return cell

        df = df.map(retain_suffix)
        df.columns = list(map(retain_suffix, df.columns))

        return pd.DataFrame(
            {
                "ICD9": cls.extract_icd9_codes(df),
                "NODE_IDX": list(df["Class ID"]),
                "PARENT_IDX": list(df["Parents"]),
                "LABEL": list(df["Preferred Label"]),
            }
        )

    @classmethod
    def parent_child_mappings(cls, df: pd.DataFrame) -> dict[str, frozenset[str]]:
        pt2ch = df.groupby("PARENT_IDX")["NODE_IDX"].apply(frozenset).to_dict()
        # Remove dummy parent of diagnoses.
        del pt2ch[cls.DUMMY_ROOT_CLASS_ID]
        return pt2ch

    @staticmethod
    def generate_dictionaries(df: pd.DataFrame) -> dict[str, Any]:
        # df version for leaf nodes only (with non-empty ICD9 codes)
        df_leaves = df[df["ICD9"] != ""]

        icd2dag = dict(zip(df_leaves["ICD9"], df_leaves["NODE_IDX"]))

        # df version for internal nodes only (with empty ICD9 codes)
        df_internal = df[(df["ICD9"] == "") | df["ICD9"].isnull()]

        icd_codes = sorted(df_leaves["ICD9"])
        icd_desc = dict(zip(df_leaves["ICD9"], df_leaves["LABEL"]))

        dag_codes = list(map(icd2dag.get, icd_codes)) + sorted(df_internal["NODE_IDX"])
        dag_desc = dict(zip(df["NODE_IDX"], df["LABEL"]))

        return {
            "codes": tuple(sorted(icd_codes)),
            "desc": FrozenDict11(icd_desc),
            "code2dag": FrozenDict11(icd2dag),
            "dag_codes": tuple(sorted(dag_codes)),
            "dag_desc": FrozenDict11(dag_desc),
        }

    @classmethod
    def create_scheme(cls) -> ICD9PCS:
        # to reduce time of redundant processing.
        processed_icd_table = cls.process_icd_table(cls.load_raw_table())
        all_parent_to_children_map = cls.parent_child_mappings(processed_icd_table)
        return ICD9PCS(name="icd9pcs", **cls.create_pcs_scheme_data(processed_icd_table, all_parent_to_children_map))


class ICD9CMFactory(ICD9PCSFactory):
    ICD9_FILE: Final[str] = resources_path("ICD", "ICD9CM.csv.gz")
    DUMMY_ROOT_CLASS_ID: Final[str] = "owl#Thing"
    PCS_ROOT_CLASS_ID: Final[str] = "00-99.99"
    CM_ROOT_CLASS_ID: Final[str] = "001-999.99"

    @classmethod
    def extract_icd9_codes(cls, df: pd.DataFrame) -> list[str]:
        return df["Class ID"].tolist()

    @classmethod
    def create_scheme(cls) -> ICD9CM:
        # to reduce time of redundant processing.
        processed_icd_table = cls.process_icd_table(cls.load_raw_table())
        all_parent_to_children_map = cls.parent_child_mappings(processed_icd_table)
        return ICD9CM(name="icd9cm", **cls.create_cm_scheme_data(processed_icd_table, all_parent_to_children_map))



from copy import deepcopy
from typing import cast
from unittest import mock

import ehrax as rx
import equinox as eqx
import numpy as np
import pytest
import tables as tb
from .common_setup import ADMISSION_CONCEPT_MAX_STAY_HOURS
from ehrax.testing.common_setup import (
    BINARY_OBSERVATION_CODE_INDEX,
    CATEGORICAL_OBSERVATION_CODE_INDEX,
    inpatient_binary_input,
    leading_observables_extractor,
    NUMERIC_OBSERVATION_CODE_INDEX,
    ORDINAL_OBSERVATION_CODE_INDEX,
    SCHEMES,
)


class TestInpatientObservables:
    def test_empty(self):
        obs = rx.InpatientObservables.empty(len(SCHEMES["obs"]))
        assert len(obs.time) == 0
        assert len(obs.value) == 0
        assert len(obs.mask) == 0
        assert all(a.shape[:2] == (0, len(SCHEMES["obs"])) for a in [obs.value, obs.mask])

    @pytest.mark.parametrize("time_valid_dtype", [np.float64])
    @pytest.mark.parametrize("mask_valid_dtype", [bool])
    @pytest.mark.parametrize("time_invalid_dtype", [np.int32, np.int64, np.float32, bool, str, object])
    @pytest.mark.parametrize("mask_invalid_dtype", [np.int32, np.int64, np.float32, str, object])
    def test_invalid_init(self, time_invalid_dtype, mask_invalid_dtype, time_valid_dtype, mask_valid_dtype):
        obs = rx.InpatientObservables(
            time=np.array([1.0, 2.0], dtype=time_valid_dtype),
            value=np.array([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]]),
            mask=np.array([[True, False, True], [False, True, False]], dtype=mask_valid_dtype),
        )

        # Invalid value shape
        with pytest.raises(AssertionError):
            rx.InpatientObservables(time=obs.time[:1], value=obs.value[:1].squeeze(), mask=obs.mask[:1])

        # Invalid mask shape
        with pytest.raises(AssertionError):
            rx.InpatientObservables(time=obs.time[:1], value=obs.value[:1], mask=obs.mask[:1].squeeze())

        # inconsistent time and value shape
        with pytest.raises(AssertionError):
            rx.InpatientObservables(time=obs.time, value=obs.value[:-1], mask=obs.mask)

        # inconsistent time and mask shape
        with pytest.raises(AssertionError):
            rx.InpatientObservables(time=obs.time, value=obs.value, mask=obs.mask[:-1])

        with pytest.raises(AssertionError):
            rx.InpatientObservables(
                time=np.array([1.0, 2.0], dtype=time_invalid_dtype),
                value=np.array([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]]),
                mask=np.array([[True, False, True], [False, True, False]], dtype=mask_valid_dtype),
            )

        with pytest.raises(AssertionError):
            rx.InpatientObservables(
                time=np.array([1.0, 2.0], dtype=time_valid_dtype),
                value=np.array([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]]),
                mask=np.array(
                    [[True, False, True], [False, True, False]],
                    dtype=mask_invalid_dtype,
                ),
            )

    def test_time_sorted(self, inpatient_observables: rx.InpatientObservables):
        assert np.all(np.diff(inpatient_observables.time) >= 0)

    def test_len(self, inpatient_observables: rx.InpatientObservables):
        assert len(inpatient_observables) == len(inpatient_observables.time)

    def test_equal(self, inpatient_observables: rx.InpatientObservables):
        assert inpatient_observables.equals(deepcopy(inpatient_observables))
        if len(inpatient_observables) == 0:
            raise pytest.skip("No observations to test")
        time = inpatient_observables.time.copy()
        time[-1] = time[-1] + 1
        c1 = eqx.tree_at(lambda x: x.time, inpatient_observables, time)
        assert not inpatient_observables.equals(c1)

        value = inpatient_observables.value.copy()
        value[-1, -1] = np.nan
        c2 = eqx.tree_at(lambda x: x.value, inpatient_observables, value)
        assert not inpatient_observables.equals(c2)

        mask = inpatient_observables.mask.astype(int)
        c3 = eqx.tree_at(lambda x: x.mask, inpatient_observables, mask)
        assert not inpatient_observables.equals(c3)

    def test_as_dataframe(self):
        pass

    def test_groupby_code(self):
        pass

    @pytest.fixture(
        params=[np.array([5.0]), np.array([0.5, 2.0]), np.array([1.0, 3.0, 5.0])],
        scope="class",
    )
    def sep(self, request):
        return request.param

    @pytest.fixture(scope="class")
    def segmented_inpatient_observables(self, inpatient_observables: rx.InpatientObservables, sep):
        return rx.SegmentedInpatientObservables.from_observables(inpatient_observables, sep)

    def test_segmentation_concat(
        self,
        inpatient_observables: rx.InpatientObservables,
        segmented_inpatient_observables: rx.SegmentedInpatientObservables,
        sep: np.ndarray,
    ):
        seg = segmented_inpatient_observables.segments
        assert len(seg) == len(sep) + 1
        assert sum(len(s) for s in seg) == len(inpatient_observables)
        assert sum(s.time.size + s.value.size + s.mask.size for s in seg) == (
            inpatient_observables.time.size + inpatient_observables.value.size + inpatient_observables.mask.size
        )
        assert inpatient_observables.equals(rx.InpatientObservables.concat(seg))
        assert all(seg[i].time.max() <= sep[i] for i in range(len(sep)) if len(seg[i]) > 0)
        assert all(seg[i + 1].time.min() >= sep[i] for i in range(len(sep)) if len(seg[i + 1]) > 0)

    def test_hf5_group_serialization(self, inpatient_observables: rx.InpatientObservables, hf5_group_writer: tb.Group):
        inpatient_observables.to_hdf_group(hf5_group_writer)
        assert inpatient_observables.equals(rx.InpatientObservables.from_hdf_group(hf5_group_writer))

    def test_hd5_group_serialization_segmented(
        self,
        segmented_inpatient_observables: rx.SegmentedInpatientObservables,
        hf5_group_writer: tb.Group,
    ):
        segmented_inpatient_observables.to_hdf_group(hf5_group_writer)
        assert segmented_inpatient_observables.equals(rx.SegmentedInpatientObservables.from_hdf_group(hf5_group_writer))

    @pytest.mark.parametrize("ntype", ["N", "B", "C", "O"])
    def test_type_aggregator(self, ntype: rx.NumericalTypeHint):
        a = rx.InpatientObservables.agg(ntype, np.array([1, 2, 3]).reshape(-1, 1, 1), np.ones((3,), dtype=bool))
        assert np.size(a) == 1

    # mask = (1, 0, 1)
    # @pytest.mark.serial_test
    @pytest.mark.parametrize(
        "x,ntype,out",
        [
            (np.array([1.0, 2.0, 3.0]), ("N",), 2.0),
            (np.array([1, 0, 1]), ("B",), 1.0),
            (np.array([0, 1, 0]), ("B",), 0.0),
            (np.array([0, 1, 0]), ("C",), 0),
            (np.array([1, 0, 1]), ("C",), 1),
            (np.array([0, 1, 0]), ("O",), 0),
            (np.array([1, 0, 1]), ("O",), 1),
            (np.array([1, 0, 0]), ("O",), 1),
            (
                np.array([[1.0, 2.0, 3.0], [2.0, 4.0, 5.0], [2.0, 1.0, 6.0]]),
                ("N", "N", "N"),
                np.array([1.5, 1.5, 4.5]),
            ),
            (
                np.array(
                    [
                        [[1.0, 0.0], [2.0, 1.0], [3.0, 0.0]],
                        [[2.0, 4.0], [4.0, 0.0], [5.0, 1.0]],
                        [[6.0, 1.0], [7.0, 0.0], [8.0, 0.0]],
                    ]
                ),
                ("N", "N", "N"),
                np.array([[3.5, 0.5], [4.5, 0.5], [5.5, 0.0]]),
            ),
        ],
    )
    def test_time_binning_aggregate(self, x, ntype: tuple[rx.NumericalTypeHint, ...], out):
        if x.ndim == 1:
            x = x.reshape(-1, 1, 1)
        if x.ndim == 2:
            x = x.reshape(x.shape + (1,))

        time_mask = np.broadcast_to(np.array([1, 0, 1]).reshape(-1, 1).astype(bool), x.shape[:2])
        out = np.array([out]).reshape((1,) + x.shape[1:])

        np.testing.assert_equal(rx.InpatientObservables._time_binning_aggregate(x, time_mask, ntype)[0], out)
        # if the input mask is all zeros, then the resulting mask is all zeros:
        v, m = rx.InpatientObservables._time_binning_aggregate(x, np.zeros_like(time_mask, dtype=bool), ntype)
        assert all(m == 0)

        with pytest.raises(AssertionError):
            rx.InpatientObservables._time_binning_aggregate(x, time_mask.astype(int), ntype)

        if any(d == 1 for d in x.shape):
            with pytest.raises(AssertionError):
                rx.InpatientObservables._time_binning_aggregate(x.squeeze(), time_mask, ntype)

    @pytest.mark.parametrize("hours", [1.0, 2.0, 3.0])
    def test_time_binning(self, hours: float, inpatient_observables: rx.InpatientObservables):
        if len(inpatient_observables) == 0:
            raise pytest.skip("No observations to test")

        binned = inpatient_observables.time_binning(hours, SCHEMES["obs"].types)
        assert np.all(binned.time % hours == 0.0)
        assert sorted(binned.time) == binned.time.tolist()

        # test causality. inject np.inf to the time array and ensure that
        # values in future time bins do not leak into the past ones.
        for i, ti in enumerate(inpatient_observables.time):
            val = inpatient_observables.value.copy()
            # NUMERIC_OBSERVATION_CODE_INDEX is aggregated with mean.
            val[i:, NUMERIC_OBSERVATION_CODE_INDEX] = np.inf
            obs = eqx.tree_at(lambda x: x.value, inpatient_observables, val)
            binned = obs.time_binning(hours, SCHEMES["obs"].types)
            assert np.all(binned.value[binned.time < ti] < np.inf)

            mask = inpatient_observables.mask.copy()
            if mask[i:, NUMERIC_OBSERVATION_CODE_INDEX].sum() > 0:
                assert np.any(np.isinf(binned.value[binned.time >= ti]))


class TestLeadingObservableExtractor:
    @pytest.mark.parametrize("leading_hours", [[1.0], [2.0], [1.0, 2.0], [1.0, 2.0, 3.0]])
    def test_len(self, leading_hours: list[float]):
        extractor = leading_observables_extractor(
            observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]), leading_hours=leading_hours
        )
        assert len(extractor) == len(leading_hours)

    @pytest.mark.parametrize("leading_hours", [[1.0], [2.0], [1.0, 2.0], [1.0, 2.0, 3.0]])
    @pytest.mark.parametrize("entry_neglect_window", [0.0, 1.0, 2.0])
    @pytest.mark.parametrize("recovery_window", [0.0, 1.0, 2.0])
    @pytest.mark.parametrize("minimum_acquisitions", [0, 1, 2, 3])
    def test_init(
        self,
        leading_hours: list[float],
        entry_neglect_window: float,
        recovery_window: float,
        minimum_acquisitions: int,
    ):
        if len(leading_hours) < 2:
            raise pytest.skip("Not enough leading hours to test")

        with pytest.raises(AssertionError):
            # leading hours must be sorted
            leading_observables_extractor(
                observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]),
                leading_hours=list(reversed(leading_hours)),
                entry_neglect_window=entry_neglect_window,
                recovery_window=recovery_window,
                minimum_acquisitions=minimum_acquisitions,
                code_index=BINARY_OBSERVATION_CODE_INDEX,
            )

        with pytest.raises(AssertionError):
            # categorical and numerical codes are not supported, yet.
            leading_observables_extractor(
                observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]),
                leading_hours=leading_hours,
                entry_neglect_window=entry_neglect_window,
                recovery_window=recovery_window,
                minimum_acquisitions=minimum_acquisitions,
                code_index=CATEGORICAL_OBSERVATION_CODE_INDEX,
            )

        with pytest.raises(AssertionError):
            # categorical and numerical codes are not supported, yet.
            leading_observables_extractor(
                observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]),
                leading_hours=leading_hours,
                entry_neglect_window=entry_neglect_window,
                recovery_window=recovery_window,
                minimum_acquisitions=minimum_acquisitions,
                code_index=NUMERIC_OBSERVATION_CODE_INDEX,
            )

    def test_index2code(self):
        pass

    def test_index2desc(self):
        pass

    def test_code2index(self):
        pass

    @pytest.mark.parametrize("leading_hours", [[1.0], [2.0], [1.0, 2.0], [1.0, 2.0, 3.0]])
    @pytest.mark.parametrize("code_index", [BINARY_OBSERVATION_CODE_INDEX, ORDINAL_OBSERVATION_CODE_INDEX])
    def test_empty(self, leading_hours: list[float], code_index: int):
        extractor = leading_observables_extractor(
            observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]),
            leading_hours=leading_hours,
            code_index=code_index,
        )
        empty = extractor.empty()
        assert len(extractor) == len(leading_hours)
        assert empty.value.shape[1] == len(extractor)
        assert empty.mask.shape[1] == len(extractor)

        assert len(empty) == 0
        assert empty.time.size == 0
        assert empty.value.size == 0
        assert empty.mask.size == 0

    @pytest.mark.parametrize(
        "x, y",
        [
            (
                np.array([1.0, 2.0, 3.0]),
                np.array([[1.0, 2.0, 3.0], [2.0, 3.0, np.nan], [3.0, np.nan, np.nan]]),
            )
        ],
    )
    def test_nan_concat_leading_windows(self, x, y):
        assert np.array_equal(
            rx.LeadingObservableExtractor._nan_concat_leading_windows(x),
            y,
            equal_nan=True,
        )

    @pytest.mark.parametrize(
        "x, y",
        [
            (
                np.array([[1.0, 2.0, 3.0], [2.0, 3.0, np.nan], [3.0, np.nan, np.nan]]),
                np.array([1.0, 1.0, 1.0]),
            ),
            (
                np.array([[1.0, 2.0, -3.0], [0.0, 0.0, np.nan], [np.nan, np.nan, np.nan]]),
                np.array([1.0, 0.0, np.nan]),
            ),
        ],
    )
    def test_nan_agg_nonzero(self, x, y):
        assert np.array_equal(rx.LeadingObservableExtractor._nan_agg_nonzero(x, axis=1), y, equal_nan=True)

    @pytest.mark.parametrize(
        "x, y",
        [
            (
                np.array([[1.0, 2.0, 3.0], [2.0, 3.0, np.nan], [3.0, np.nan, np.nan]]),
                np.array([3.0, 3.0, 3.0]),
            ),
            (
                np.array([[1.0, 2.0, -3.0], [0.0, 0.0, np.nan], [np.nan, np.nan, np.nan]]),
                np.array([2.0, 0.0, np.nan]),
            ),
        ],
    )
    def test_nan_agg_max(self, x, y):
        assert np.array_equal(rx.LeadingObservableExtractor._nan_agg_max(x, axis=1), y, equal_nan=True)

    @pytest.mark.parametrize("minimum_acquisitions", [0, 1, 2, 3])
    def test_filter_first_acquisitions(self, inpatient_observables: rx.InpatientObservables, minimum_acquisitions: int):
        lead_extractor = leading_observables_extractor(
            observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]), minimum_acquisitions=minimum_acquisitions
        )
        m = lead_extractor.filter_first_acquisitions(len(inpatient_observables), minimum_acquisitions)
        n_affected = min(minimum_acquisitions, len(inpatient_observables))
        assert (~m).sum() == n_affected
        assert (~m)[n_affected:].sum() == 0

    @pytest.mark.parametrize("entry_neglect_window", [0.0, 1.0, 2.0])
    def test_neutralize_entry_neglect_window(
        self, inpatient_observables: rx.InpatientObservables, entry_neglect_window: int
    ):
        lead_extractor = leading_observables_extractor(
            observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]), entry_neglect_window=entry_neglect_window
        )
        t = inpatient_observables.time
        m = lead_extractor.filter_entry_neglect_window(t, entry_neglect_window)
        n_affected = np.sum(t <= entry_neglect_window)
        assert (~m).sum() == n_affected
        assert (~m)[n_affected:].sum() == 0

    @pytest.mark.parametrize("recovery_window", [0.0, 1.0, 2.0])
    def test_neutralize_recovery_window(self, inpatient_observables: rx.InpatientObservables, recovery_window: float):
        if len(inpatient_observables) == 0:
            raise pytest.skip("No observations to test")

        lead_extractor = leading_observables_extractor(
            observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]), entry_neglect_window=recovery_window
        )
        x = inpatient_observables.value[:, lead_extractor.code_index]
        t = inpatient_observables.time
        m = lead_extractor.filter_recovery_window(t, x, recovery_window)

        arg_x_recovery = np.flatnonzero((x[:-1] != 0) & (x[1:] == 0))
        assert np.isnan(x).sum() == 0
        for i, arg in enumerate(arg_x_recovery):
            last = len(t)
            if i < len(arg_x_recovery) - 1:
                last = arg_x_recovery[i + 1]

            ti = t[arg + 1 : last] - t[arg]
            mi = m[arg + 1 : last]
            n_affected = np.sum(ti <= recovery_window)

            assert (~mi).sum() == n_affected

            assert (~mi)[n_affected:].sum() == 0

    @pytest.mark.parametrize("minimum_acquisitions", [0, 100])
    @pytest.mark.parametrize("entry_neglect_window", [0.0, 1000.0])
    @pytest.mark.parametrize("recovery_window", [0.0, 10000.0])
    def test_mask_noisy_observations(
        self,
        inpatient_observables: rx.InpatientObservables,
        minimum_acquisitions: int,
        entry_neglect_window: int,
        recovery_window: float,
    ):
        lead_extractor = leading_observables_extractor(
            observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]),
            minimum_acquisitions=minimum_acquisitions,
            entry_neglect_window=entry_neglect_window,
            recovery_window=recovery_window,
        )
        x = inpatient_observables.value[:, lead_extractor.code_index]
        t = inpatient_observables.time
        qual = f"{lead_extractor.__module__}.{type(lead_extractor).__qualname__}"
        with mock.patch(f"{qual}.filter_first_acquisitions") as mocker1:
            with mock.patch(f"{qual}.filter_entry_neglect_window") as mocker2:
                with mock.patch(f"{qual}.filter_recovery_window") as mocker3:
                    lead_extractor.mask_noisy_observations(
                        t,
                        x,
                        minimum_acquisitions=minimum_acquisitions,
                        entry_neglect_window=entry_neglect_window,
                        recovery_window=recovery_window,
                    )
                    mocker1.assert_called_once_with(len(t), minimum_acquisitions)
                    mocker2.assert_called_once_with(t, entry_neglect_window)
                    mocker3.assert_called_once_with(t, x, recovery_window)

    @pytest.mark.parametrize("code_index", [BINARY_OBSERVATION_CODE_INDEX, ORDINAL_OBSERVATION_CODE_INDEX])
    def test_extract_leading_window(self, inpatient_observables: rx.InpatientObservables, code_index: int):
        lead_extractor = leading_observables_extractor(
            observation_scheme=cast(rx.NumericScheme, SCHEMES["obs"]), code_index=code_index
        )
        x = inpatient_observables.value[:, lead_extractor.code_index, :]
        m = inpatient_observables.mask[:, lead_extractor.code_index]
        t = inpatient_observables.time
        lead = lead_extractor(inpatient_observables)

        assert len(lead) == len(inpatient_observables)
        assert lead.value.shape[1] == len(lead_extractor)

        for iw, w in enumerate(lead_extractor.config.leading_hours):
            for i in range(len(t)):
                delta_t = t[i:] - t[i]
                indexer = (delta_t <= w) & m[i:]
                xi = x[i:][indexer]
                yi = lead.value[i, iw]
                # if all are nan, then the lead is nan.
                if np.size(xi) == 0:
                    assert np.isnan(yi)
                elif code_index == BINARY_OBSERVATION_CODE_INDEX:
                    assert yi == xi.any()
                elif code_index == ORDINAL_OBSERVATION_CODE_INDEX:
                    assert yi == xi.max()
                else:
                    assert False

    def test_apply(self):
        pass


class TestInpatientInput:
    def test_init(self):
        pass

    def test_apply(self):
        # TODO: priority.
        pass

    @pytest.mark.parametrize("n", [0, 1, 100, 501])
    @pytest.mark.parametrize("p", [1, 100, 501])
    def test_hf5_group_serialization(self, n: int, p: int, hf5_group_writer: tb.Group):
        input = inpatient_binary_input(n, p, los_hours=ADMISSION_CONCEPT_MAX_STAY_HOURS)
        input.to_hdf_group(hf5_group_writer)
        loaded_input = rx.InpatientInput.from_hdf_group(hf5_group_writer)
        assert len(input) == len(loaded_input)
        assert input.equals(loaded_input)

    def test_empty(self):
        pass


class TestInpatientInterventions:
    def test_hf5_group_serialization(
        self,
        inpatient_interventions_with_a_none: rx.InpatientInterventions,
        hf5_group_writer: tb.Group,
    ):
        inpatient_interventions_with_a_none.to_hdf_group(hf5_group_writer)
        assert inpatient_interventions_with_a_none.equals(rx.InpatientInterventions.from_hdf_group(hf5_group_writer))

    def test_timestamps(self, inpatient_interventions_with_a_none: rx.InpatientInterventions):
        timestamps = inpatient_interventions_with_a_none.timestamps
        assert len(timestamps) == len(set(timestamps))
        assert all(isinstance(t, float) for t in timestamps)
        assert list(sorted(timestamps)) == timestamps
        ts = set()
        if inpatient_interventions_with_a_none.hosp_procedures is not None:
            ts.update(inpatient_interventions_with_a_none.hosp_procedures.starttime.tolist())
            ts.update(inpatient_interventions_with_a_none.hosp_procedures.endtime.tolist())
        if inpatient_interventions_with_a_none.icu_procedures is not None:
            ts.update(inpatient_interventions_with_a_none.icu_procedures.starttime.tolist())
            ts.update(inpatient_interventions_with_a_none.icu_procedures.endtime.tolist())
        if inpatient_interventions_with_a_none.icu_inputs is not None:
            ts.update(inpatient_interventions_with_a_none.icu_inputs.starttime.tolist())
            ts.update(inpatient_interventions_with_a_none.icu_inputs.endtime.tolist())
        assert ts == set(timestamps)


class TestSegmentedInpatientInterventions:
    def test_from_inpatient_interventions(self, inpatient_interventions_with_a_none):
        assert all(
            isinstance(scheme, rx.CodingScheme)
            for scheme in (
                SCHEMES["hosp_procedures"],
                SCHEMES["icu_procedures"],
                SCHEMES["icu_inputs"],
            )
        )
        schemes = {
            "hosp_procedures": SCHEMES["hosp_procedures"],
            "icu_procedures": SCHEMES["icu_procedures"],
            "icu_inputs": SCHEMES["icu_inputs"],
        }
        seg = rx.SegmentedInpatientInterventions.from_interventions(
            inpatient_interventions_with_a_none,
            ADMISSION_CONCEPT_MAX_STAY_HOURS,
            hosp_procedures_size=len(SCHEMES["hosp_procedures"]),
            icu_procedures_size=len(SCHEMES["icu_procedures"]),
            icu_inputs_size=len(SCHEMES["icu_inputs"]),
            maximum_padding=1,
        )
        assert abs(len(seg.time) - len(inpatient_interventions_with_a_none.timestamps)) <= 2
        for k in ("hosp_procedures", "icu_procedures", "icu_inputs"):
            if getattr(inpatient_interventions_with_a_none, k) is not None:
                seg_intervention = getattr(seg, k)
                assert seg_intervention.shape == (len(seg.time) - 1, len(schemes[k]))

            else:
                assert getattr(seg, k) is None

    @pytest.mark.parametrize("array", [np.array([1.0, 2.0, 3.0]), np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])])
    @pytest.mark.parametrize("value", [0.0, np.nan])
    @pytest.mark.parametrize("maximum_padding", [1, 2, 3, 5, 10, 100])
    def test_pad_array(self, array, value, maximum_padding):
        y = rx.SegmentedInpatientInterventions.pad_array(array, value=value, maximum_padding=maximum_padding)
        assert len(y) >= len(array)
        assert len(y) <= len(array) + maximum_padding
        assert np.all(y[: len(array)] == array)
        if np.isnan(value):
            assert np.isnan(y[len(array) :]).all()
        else:
            assert np.all(y[len(array) :] == value)

        if len(array) == maximum_padding or maximum_padding == 1:
            assert len(y) == len(array)

    @pytest.mark.parametrize("test_target", ["hosp_procedures", "icu_procedures", "icu_inputs"])
    def test_segmentation(
        self,
        inpatient_interventions_with_a_none: rx.InpatientInterventions,
        test_target: str,
    ):
        inpatient_intervention = getattr(inpatient_interventions_with_a_none, test_target)
        if inpatient_intervention is None or inpatient_intervention.starttime.size == 0:
            raise pytest.skip("No interventions to test")

        scheme = {
            "hosp_procedures": SCHEMES["hosp_procedures"],
            "icu_procedures": SCHEMES["icu_procedures"],
            "icu_inputs": SCHEMES["icu_inputs"],
        }[test_target]
        seg = rx.SegmentedInpatientInterventions._segment(
            inpatient_intervention.starttime, inpatient_intervention, len(scheme)
        )
        for i, t in enumerate(inpatient_intervention.starttime):
            assert np.array_equal(inpatient_intervention(t, len(scheme)), seg[i])

    def test_hf5_group_serialization(
        self,
        segmented_inpatient_interventions: rx.SegmentedInpatientInterventions,
        hf5_group_writer: tb.Group,
    ):
        segmented_inpatient_interventions.to_hdf_group(hf5_group_writer)
        assert segmented_inpatient_interventions.equals(
            rx.SegmentedInpatientInterventions.from_hdf_group(hf5_group_writer)
        )


class TestAdmission:
    def test_hf5_group_serialization(self, admission: rx.Admission, hf5_group_writer: tb.Group):
        admission.to_hdf_group(hf5_group_writer)
        assert admission.equals(rx.Admission.from_hdf_group(hf5_group_writer))

    def test_interval_hours(self):
        pass

    def test_interval_days(self):
        pass

    def test_days_since(self):
        pass


class TestSegmentedAdmission:
    def test_from_admission(self, admission: rx.Admission, segmented_admission: rx.SegmentedAdmission):
        assert segmented_admission.admission_id == admission.admission_id
        assert segmented_admission.admission_dates == admission.admission_dates
        assert segmented_admission.dx_codes == admission.dx_codes
        assert segmented_admission.dx_codes_history == admission.dx_codes_history
        assert segmented_admission.outcome == admission.outcome
        assert segmented_admission.observables is not None
        assert segmented_admission.leading_observable is not None
        if admission.observables is not None:
            assert rx.InpatientObservables.concat(segmented_admission.observables.segments).equals(
                admission.observables
            )
        if admission.leading_observable is not None:
            assert rx.InpatientObservables.concat(segmented_admission.leading_observable.segments).equals(
                admission.leading_observable
            )
        if admission.interventions is None:
            return
        interventions = admission.interventions
        seg_interventions = segmented_admission.interventions
        timestamps = admission.interventions.timestamps

        time = sorted(set([0.0] + timestamps + [ADMISSION_CONCEPT_MAX_STAY_HOURS]))
        for i, (start, end) in enumerate(zip(time[:-1], time[1:])):
            if interventions.hosp_procedures is not None:
                assert seg_interventions is not None and seg_interventions.hosp_procedures is not None
                p = seg_interventions.hosp_procedures.shape[1]
                hosp_procedure = interventions.hosp_procedures(start, p)
                seg_hosp_procedure = seg_interventions.hosp_procedures[i]
                assert np.array_equal(seg_hosp_procedure, hosp_procedure)
            if interventions.icu_procedures is not None:
                assert seg_interventions is not None and seg_interventions.icu_procedures is not None
                p = seg_interventions.icu_procedures.shape[1]
                icu_procedure = interventions.icu_procedures(start, p)
                seg_icu_procedure = seg_interventions.icu_procedures[i]
                assert np.array_equal(seg_icu_procedure, icu_procedure)
            if interventions.icu_inputs is not None:
                assert seg_interventions is not None and seg_interventions.icu_inputs is not None
                p = seg_interventions.icu_inputs.shape[1]
                icu_input = interventions.icu_inputs(start, p)
                seg_icu_input = seg_interventions.icu_inputs[i]
                assert np.array_equal(seg_icu_input, icu_input)

            if admission.observables is not None:
                obs_time = admission.observables.time
                segment_time = segmented_admission.observables[i].time
                assert np.all((start <= segment_time) & (segment_time <= end))
                if i == len(time) - 2:
                    time_mask = (obs_time >= start) & (obs_time <= end)
                else:
                    time_mask = (obs_time >= start) & (obs_time < end)
                segment_val = segmented_admission.observables[i].value
                segment_mask = segmented_admission.observables[i].mask
                val = admission.observables.value[time_mask]
                mask = admission.observables.mask[time_mask]
                assert np.array_equal(segment_val, val)
                assert np.array_equal(segment_mask, mask)

    def test_hf5_group_serialization(self, segmented_admission: rx.SegmentedAdmission, hf5_group_writer: tb.Group):
        segmented_admission.to_hdf_group(hf5_group_writer)
        assert segmented_admission.equals(rx.SegmentedAdmission.from_hdf_group(hf5_group_writer))


class TestStaticInfo:
    def test_hf5_group_serialization(self, static_info: rx.StaticInfo, hf5_group_writer: tb.Group):
        static_info.to_hdf_group(hf5_group_writer)
        assert static_info.equals(rx.StaticInfo.from_hdf_group(hf5_group_writer))

    def test_constant_attributes(self):
        pass

    def test_constant_vec(self):
        pass

    def test_dynamic_attributes(self):
        pass


class TestPatient:
    def test_d2d_interval_days(self):
        pass

    def test_outcome_frequency(self):
        pass

    def test_hf5_group_serialization(self, patient: rx.Patient, tmpdir: str):
        path = f"{tmpdir}/patient.h5"
        with tb.open_file(path, "w") as f:
            patient.to_hdf_group(f.create_group("/", "patient"))

        with tb.open_file(path, "r") as f:
            loaded_patient = rx.Patient.from_hdf_group(f.root.patient)
            assert patient.equals(loaded_patient)


class TestSegmentedPatient:
    def test_d2d_interval_days(self):
        pass

    def test_outcome_frequency(self):
        pass

    def test_hf5_group_serialization(self, segmented_patient: rx.SegmentedPatient, tmpdir: str):
        path = f"{tmpdir}/segmented_patient.h5"
        with tb.open_file(path, "w") as f:
            segmented_patient.to_hdf_group(f.create_group("/", "seg_patient"))

        with tb.open_file(path, "r") as f:
            loaded_segmented_patient = rx.SegmentedPatient.from_hdf_group(f.root.seg_patient)
            assert segmented_patient.equals(loaded_segmented_patient)



import ehrax as rx
import equinox as eqx
import numpy as np
import pandas as pd
import pytest
from ehrax.testing.common_setup import DATASET_CONFIG, DATASET_SCHEME_MANAGER, SCHEMES

from .conftest import Dataset


@pytest.fixture(scope="module")
def mimiciv_dataset_without_uom_normalization(
    dataset_tables_with_records: rx.DatasetTables, unit_converter_table: pd.DataFrame
) -> Dataset:
    config = eqx.tree_at(
        lambda x: x.scheme, DATASET_CONFIG, rx.DatasetSchemeConfig(**DATASET_CONFIG.scheme.scheme_fields())
    )
    ds = Dataset(tables=dataset_tables_with_records, config=config)
    return ds._execute_pipeline([rx.SetIndex(), rx.SynchronizeSubjects(), rx.CastTimestamps()], DATASET_SCHEME_MANAGER)


class TestUnitConversionAndFilterInvalidInputRates:
    @pytest.fixture(scope="class")
    def fixed_dataset(self, mimiciv_dataset_without_uom_normalization: Dataset) -> Dataset:
        return rx.ICUInputRateUnitConversion.apply(
            mimiciv_dataset_without_uom_normalization, DATASET_SCHEME_MANAGER, rx.Report()
        )[0]

    @pytest.fixture(scope="class")
    def icu_inputs_unfixed(self, mimiciv_dataset_without_uom_normalization: Dataset):
        return mimiciv_dataset_without_uom_normalization.tables.icu_inputs

    @pytest.fixture(scope="class")
    def icu_inputs_fixed(self, fixed_dataset: Dataset):
        return fixed_dataset.tables.icu_inputs

    @pytest.fixture(scope="class")
    def derived_icu_inputs_cols(self):
        assert DATASET_CONFIG.columns.icu_inputs is not None, (
            "Dataset configuration does not have icu_inputs table defined."
        )
        c = DATASET_CONFIG.columns.icu_inputs
        return [
            c.derived_unit_normalization_factor,
            c.derived_universal_unit,
            c.derived_normalized_amount,
            c.derived_normalized_amount_per_hour,
        ]

    def test_icu_input_rate_unit_conversion(
        self, icu_inputs_fixed: pd.DataFrame, icu_inputs_unfixed: pd.DataFrame, derived_icu_inputs_cols: list[str]
    ):
        assert all(c not in icu_inputs_unfixed.columns for c in derived_icu_inputs_cols)
        assert all(c in icu_inputs_fixed.columns for c in derived_icu_inputs_cols)
        assert DATASET_CONFIG.columns.icu_inputs is not None, (
            "Dataset configuration does not have icu_inputs table defined."
        )
        c = DATASET_CONFIG.columns.icu_inputs

        scheme: rx.CodingSchemeWithUOM = SCHEMES["icu_inputs"]
        # For every (code, unit) pair, a unique normalization factor and universal unit is assigned.
        for (code, unit), inputs_df in icu_inputs_fixed.groupby([c.code, c.amount_unit]):
            assert inputs_df[c.derived_universal_unit].unique() == scheme.universal_unit[code]
            assert (
                inputs_df[c.derived_unit_normalization_factor].unique() == scheme.uom_normalization_factor[code][unit]
            )
            assert inputs_df[c.derived_normalized_amount].equals(
                inputs_df[c.amount] * scheme.uom_normalization_factor[code][unit]
            )

    @pytest.fixture(scope="class")
    def nan_inputs_dataset(self, fixed_dataset: pd.DataFrame):
        icu_inputs = fixed_dataset.tables.icu_inputs.copy()
        assert DATASET_CONFIG.columns.icu_inputs is not None, (
            "Dataset configuration does not have icu_inputs table defined."
        )
        c = DATASET_CONFIG.columns.icu_inputs
        admission_id = icu_inputs.iloc[0][c.admission_id]
        icu_inputs.loc[icu_inputs[c.admission_id] == admission_id, c.derived_normalized_amount_per_hour] = np.nan
        return eqx.tree_at(lambda x: x.tables.icu_inputs, fixed_dataset, icu_inputs)

    @pytest.fixture(scope="class")
    def filtered_dataset(self, nan_inputs_dataset: Dataset):
        return rx.FilterInvalidInputRatesSubjects.apply(nan_inputs_dataset, DATASET_SCHEME_MANAGER, rx.Report())[0]

    def test_filter_invalid_input_rates_subjects(self, nan_inputs_dataset: Dataset, filtered_dataset: Dataset):
        icu_inputs0 = nan_inputs_dataset.tables.icu_inputs
        admissions0 = nan_inputs_dataset.tables.admissions
        static0 = nan_inputs_dataset.tables.static

        icu_inputs1 = filtered_dataset.tables.icu_inputs
        admissions1 = filtered_dataset.tables.admissions
        static1 = filtered_dataset.tables.static
        assert DATASET_CONFIG.columns.icu_inputs is not None, (
            "Dataset configuration does not have icu_inputs table defined."
        )
        assert DATASET_CONFIG.columns.admissions is not None, (
            "Dataset configuration does not have admissions table defined."
        )
        assert icu_inputs0 is not None, "ICU inputs table is not present in the dataset."
        assert icu_inputs1 is not None, "Filtered ICU inputs table is not present in the dataset."
        cicu = DATASET_CONFIG.columns.icu_inputs
        cadm = DATASET_CONFIG.columns.admissions

        admission_id = icu_inputs0.iloc[0][cicu.admission_id]
        subject_id = static0[static0.index == admissions0.loc[admission_id, cadm.subject_id]].index[0]
        subject_admissions = admissions0[admissions0[cadm.subject_id] == subject_id]

        assert subject_id not in static1.index
        assert not subject_admissions.index.isin(admissions1.index).all()
        assert not subject_admissions.index.isin(icu_inputs1[cicu.admission_id]).all()
        assert icu_inputs0[cicu.derived_normalized_amount_per_hour].isna().any()
        assert not icu_inputs1[cicu.derived_normalized_amount_per_hour].isna().any()



import logging
from collections import defaultdict
from typing import Final

import pandas as pd

from ..coding_scheme import CodeMap, CodingScheme, CodingSchemesManager, FrozenDict1N, FrozenDict11, HierarchicalScheme
from ..utils import dataframe_log, resources_path
from .icd import ICDScheme
from .icd9 import ICD9CM, ICD9PCS
from .icd10 import ICD10CM, ICD10PCS
from .literals import (
    DxFlatCCSName,
    DxMultiCCSName,
    ICD9CMName,
    ICD9PCSName,
    ICD10CMName,
    ICD10PCSName,
    PrFlatCCSName,
    PrMultiCCSName,
)


class CommonPreprocess:  # mixin
    @staticmethod
    def raw_table(filetitle: str, skiprows: tuple[int, ...] = ()) -> pd.DataFrame:
        return pd.read_csv(resources_path("CCS", filetitle), skiprows=skiprows, dtype=str)

    @staticmethod
    def process_ccs_table(raw_ccs_table: pd.DataFrame, colmap: dict[str, str]) -> pd.DataFrame:
        df = raw_ccs_table.iloc[:, :]
        df = df[list(colmap.keys())].rename(columns=colmap)
        for c in df.columns:
            df.loc[:, c] = df.loc[:, c].str.strip("'").str.strip()
        return df

    @staticmethod
    def process_ccs_icd_table(icd_scheme: ICDScheme, ccs_table: pd.DataFrame) -> pd.DataFrame:
        ccs_table["ICD"] = ccs_table.loc[:, "ICD"].map(icd_scheme.format)
        valid_icd = ccs_table["ICD"].isin(icd_scheme.codes)
        unsupported_icd = ccs_table.loc[~valid_icd, :]
        dataframe_log.info(
            f"In processing CCS multi-level table mapping to {icd_scheme.name} "
            f"{(~valid_icd).sum()} (of {len(valid_icd)}) ICD codes were unsupported.",
            dataframe=unsupported_icd,
            tag=f"unsupported_icd_codes_by_{icd_scheme.name}",
        )
        return ccs_table.loc[valid_icd, :]


class CommonMultiLevelCCS(CommonPreprocess):  # mixin
    @staticmethod
    def c_level_code(level: int) -> str:
        return f"'CCS LVL {level}'"

    @staticmethod
    def c_level_code_description(level: int) -> str:
        return f"'CCS LVL {level} LABEL'"

    @staticmethod
    def register_multi_level_mappings(
        manager: CodingSchemesManager, ccs_scheme: str, icd_scheme: str, processed_ccs_table: pd.DataFrame
    ) -> CodingSchemesManager:
        ccs_scheme = manager.scheme[ccs_scheme]
        icd_scheme = manager.scheme[icd_scheme]
        # TODO: Do some hard coded tests for the correction of mapping.
        icd2ccs = defaultdict(set)
        ccs2icd = defaultdict(set)
        code_cols = [c for c in processed_ccs_table.columns if c.startswith("C")]
        for icd, levels in zip(processed_ccs_table["ICD"], zip(*[processed_ccs_table.loc[:, c] for c in code_cols])):
            levels = tuple(l for l in levels if l != "")
            if len(levels) > 0:
                icd2ccs[icd].add(levels[-1])
                ccs2icd[levels[-1]].add(icd)

        manager = manager.add_map(
            CodeMap(source_name=icd_scheme.name, target_name=ccs_scheme.name, data=FrozenDict1N(dict(icd2ccs)))
        )
        manager = manager.add_map(
            CodeMap(source_name=ccs_scheme.name, target_name=icd_scheme.name, data=FrozenDict1N(dict(ccs2icd)))
        )
        return manager

    @staticmethod
    def parent_child_mappings(ccs_processed_table: pd.DataFrame) -> dict[str, set[str]]:
        """Make a dictionary for parent-child connections."""
        pt2ch = {"root": set(ccs_processed_table["C1"])}
        levels = [c for c in ccs_processed_table.columns if c.startswith("C")]
        for pt_col, ch_col in zip(levels[:-1], levels[1:]):
            df_ = ccs_processed_table[(ccs_processed_table[pt_col] != "") & (ccs_processed_table[ch_col] != "")]
            df_ = df_[[pt_col, ch_col]].drop_duplicates()
            for parent_code, ch_df in df_.groupby(pt_col):
                pt2ch[parent_code] = set(ch_df[ch_col])
        return pt2ch

    @staticmethod
    def multi_level_desc_mappings(ccs_processed_table: pd.DataFrame) -> dict[str, str]:
        """Make a dictionary for CCS labels."""
        desc = {"root": "root"}
        levels = [c for c in ccs_processed_table.columns if c.startswith("C")]
        descriptions = [c for c in ccs_processed_table.columns if c.startswith("D")]
        for code_col, desc_col in zip(levels, descriptions):
            df_ = ccs_processed_table[ccs_processed_table[code_col] != ""]
            df_ = df_[[code_col, desc_col]].drop_duplicates()
            code_desc = dict(zip(df_[code_col], df_[desc_col]))
            desc.update(code_desc)
        return desc


class MultiLevelCCSICD9MapOps(CommonMultiLevelCCS):
    C_ICD_CODE: str = "'ICD-9-CM CODE'"
    DX_FILE: str = "ccs_multi_dx_tool_2015.csv.gz"
    PR_FILE: str = "ccs_multi_pr_tool_2015.csv.gz"
    DX_N_LEVELS: int = 4
    PR_N_LEVELS: int = 3

    @classmethod
    def colmap(cls, n_levels: int):
        return (
            {cls.c_level_code(l): f"C{l}" for l in range(1, n_levels + 1)}
            | {cls.c_level_code_description(l): f"D{l}" for l in range(1, n_levels + 1)}
            | {cls.C_ICD_CODE: "ICD"}
        )

    @classmethod
    def process_dx_ccs_icd_table(cls, icd_scheme: ICDScheme):
        ccs_table = cls.process_ccs_table(raw_ccs_table=cls.raw_table(cls.DX_FILE), colmap=cls.colmap(cls.DX_N_LEVELS))
        return cls.process_ccs_icd_table(icd_scheme, ccs_table)

    @classmethod
    def process_pr_ccs_icd_table(cls, icd_scheme: ICDScheme):
        ccs_table = cls.process_ccs_table(raw_ccs_table=cls.raw_table(cls.PR_FILE), colmap=cls.colmap(cls.PR_N_LEVELS))
        return cls.process_ccs_icd_table(icd_scheme, ccs_table)

    @classmethod
    def create_scheme(
        cls, name: PrMultiCCSName | DxMultiCCSName, raw_ccs_table: pd.DataFrame, n_levels: int
    ) -> HierarchicalScheme:
        df = CommonPreprocess.process_ccs_table(raw_ccs_table, colmap=cls.colmap(n_levels))
        ch2pt = HierarchicalScheme.reverse_connection(FrozenDict1N(cls.parent_child_mappings(df)))
        desc = FrozenDict11(cls.multi_level_desc_mappings(df))
        codes = tuple(sorted(desc.keys()))
        return HierarchicalScheme(name=name, ch2pt=ch2pt, codes=codes, desc=desc)

    @classmethod
    def create_dx_ccs(cls) -> HierarchicalScheme:  # expose
        return cls.create_scheme("dx_ccs", cls.raw_table(cls.DX_FILE), cls.DX_N_LEVELS)

    @classmethod
    def create_pr_ccs(cls) -> HierarchicalScheme:  # expose
        return cls.create_scheme("pr_ccs", cls.raw_table(cls.PR_FILE), cls.PR_N_LEVELS)

    @classmethod
    def register_dx_ccs_maps(
        cls, manager: CodingSchemesManager, icd_scheme: ICD9CMName
    ) -> CodingSchemesManager:  # expose
        scheme = manager.scheme[icd_scheme]
        assert isinstance(scheme, ICD9CM), f"Expected ICD9CM, got {type(scheme)}."
        return cls.register_multi_level_mappings(manager, "dx_ccs", icd_scheme, cls.process_dx_ccs_icd_table(scheme))

    @classmethod
    def register_pr_ccs_maps(
        cls, manager: CodingSchemesManager, icd_scheme: ICD9PCSName
    ) -> CodingSchemesManager:  # expose
        scheme = manager.scheme[icd_scheme]
        assert isinstance(scheme, ICD9PCS), f"Expected ICD9PCS, got {type(scheme)}."
        return cls.register_multi_level_mappings(manager, "pr_ccs", icd_scheme, cls.process_pr_ccs_icd_table(scheme))


class CommonFlatCCS(CommonPreprocess):
    @staticmethod
    def desc_flat_mappings(processed_ccs_table: pd.DataFrame) -> dict[str, str]:
        return processed_ccs_table.set_index("C")["D"].to_dict()

    @staticmethod
    def register_flat_mappings(
        manager: CodingSchemesManager,
        ccs_scheme: DxFlatCCSName | PrFlatCCSName,
        icd_scheme: str,
        processed_ccs_table: pd.DataFrame,
    ) -> CodingSchemesManager:
        ccs_scheme = manager.scheme[ccs_scheme]
        icd_scheme = manager.scheme[icd_scheme]
        # TODO: Do some hard coded tests for the correction of mapping.
        icd2ccs = processed_ccs_table.groupby("ICD")["C"].apply(set).to_dict()
        ccs2icd = processed_ccs_table.groupby("C")["ICD"].apply(set).to_dict()
        manager = manager.add_map(
            CodeMap(source_name=icd_scheme.name, target_name=ccs_scheme.name, data=FrozenDict1N(dict(icd2ccs)))
        )
        manager = manager.add_map(
            CodeMap(source_name=ccs_scheme.name, target_name=icd_scheme.name, data=FrozenDict1N(dict(ccs2icd)))
        )
        return manager


class FlatCCS2ICD9MapOps(CommonFlatCCS):
    C_ICD_CODE: Final[str] = "'ICD-9-CM CODE'"
    C_CCS_CODE: Final[str] = "'CCS CATEGORY'"
    C_CCS_DESC: Final[str] = "'CCS CATEGORY DESCRIPTION'"
    DX_FILE: Final[str] = "$dxref 2015.csv.gz"
    PR_FILE: Final[str] = "$prref 2015.csv.gz"
    skiprows: Final[tuple[int, int]] = (0, 2)

    @classmethod
    def colmap(cls):
        return {cls.C_ICD_CODE: "ICD", cls.C_CCS_CODE: "C", cls.C_CCS_DESC: "D"}

    @classmethod
    def process_dx_ccs_icd_table(cls, icd9_scheme: ICDScheme):
        return cls.process_ccs_icd_table(
            icd9_scheme, cls.process_ccs_table(cls.raw_table(cls.DX_FILE, cls.skiprows), cls.colmap())
        )

    @classmethod
    def process_pr_ccs_icd_table(cls, icd_scheme: ICDScheme):
        return cls.process_ccs_icd_table(
            icd_scheme, cls.process_ccs_table(cls.raw_table(cls.PR_FILE, cls.skiprows), cls.colmap())
        )

    @classmethod
    def create_scheme(cls, name: DxFlatCCSName | PrFlatCCSName, raw_ccs_table: pd.DataFrame) -> CodingScheme:
        desc = FrozenDict11(cls.desc_flat_mappings(cls.process_ccs_table(raw_ccs_table, cls.colmap())))
        codes = tuple(sorted(desc.keys()))
        return CodingScheme(name=name, codes=codes, desc=desc)

    @classmethod
    def create_dx_flat_ccs(cls) -> CodingScheme:  # expose
        return cls.create_scheme("dx_flat_ccs", cls.raw_table(cls.DX_FILE, cls.skiprows))

    @classmethod
    def create_pr_flat_ccs(cls) -> CodingScheme:  # expose
        return cls.create_scheme("pr_flat_ccs", cls.raw_table(cls.PR_FILE, cls.skiprows))

    @classmethod
    def register_dx_flat_ccs_maps(
        cls, manager: CodingSchemesManager, icd_scheme: ICD9CMName
    ) -> CodingSchemesManager:  # expose
        scheme = manager.scheme[icd_scheme]
        assert isinstance(scheme, ICD9CM), f"Expected ICD9CM. got {type(scheme)}."
        return cls.register_flat_mappings(manager, "dx_flat_ccs", icd_scheme, cls.process_dx_ccs_icd_table(scheme))

    @classmethod
    def register_pr_flat_ccs_maps(
        cls, manager: CodingSchemesManager, icd_scheme: ICD9PCSName
    ) -> CodingSchemesManager:  # expose
        scheme = manager.scheme[icd_scheme]
        assert isinstance(scheme, ICD9PCS), f"Expected PrHierarchicalICD9, got {type(scheme)}."
        return cls.register_flat_mappings(manager, "pr_flat_ccs", icd_scheme, cls.process_pr_ccs_icd_table(scheme))


class CCS2ICD10MapOps(CommonMultiLevelCCS):
    C_ICD_CM_CODE: Final[str] = "'ICD-10-CM CODE'"
    C_ICD_PCS_CODE: Final[str] = "'ICD-10-PCS CODE'"
    C_CCS_CODE: Final[str] = "'CCS CATEGORY'"
    C_CCS_DESC: Final[str] = "'CCS CATEGORY DESCRIPTION'"
    DX_FILE: Final[str] = "ccs_dx_icd10cm_2019_1.csv.gz"
    PR_FILE: Final[str] = "ccs_pr_icd10pcs_2020_1.csv.gz"
    DX_N_LEVELS: Final[int] = 2
    PR_N_LEVELS: Final[int] = 2

    @staticmethod
    def c_level_code(level: int) -> str:
        return f"'MULTI CCS LVL {level}'"

    @staticmethod
    def c_level_code_description(level: int) -> str:
        return f"'MULTI CCS LVL {level} LABEL'"

    @classmethod
    def dx_colmap(cls, n_levels: int) -> dict[str, str]:
        return (
            {cls.c_level_code(l): f"C{l}" for l in range(1, n_levels + 1)}
            | {cls.c_level_code_description(l): f"D{l}" for l in range(1, n_levels + 1)}
            | {cls.C_ICD_CM_CODE: "ICD", cls.C_CCS_CODE: "C", cls.C_CCS_DESC: "D"}
        )

    @classmethod
    def pr_colmap(cls, n_levels: int) -> dict[str, str]:
        colmap = cls.dx_colmap(n_levels)
        del colmap[cls.C_ICD_CM_CODE]
        colmap[cls.C_ICD_PCS_CODE] = "ICD"
        return colmap

    @classmethod
    def process_dx_ccs_icd_table(cls, icd10_scheme: ICD10CM):
        return cls.process_ccs_icd_table(
            icd10_scheme, cls.process_ccs_table(cls.raw_table(cls.DX_FILE), cls.dx_colmap(cls.DX_N_LEVELS))
        )

    @classmethod
    def process_pr_ccs_icd_table(cls, icd10_scheme: ICD10PCS):
        return cls.process_ccs_icd_table(
            icd10_scheme, cls.process_ccs_table(cls.raw_table(cls.PR_FILE), cls.pr_colmap(cls.PR_N_LEVELS))
        )

    @classmethod
    def register_dx_flat_ccs_maps(
        cls, manager: CodingSchemesManager, icd_scheme: ICD10CMName
    ) -> CodingSchemesManager:  # expose
        scheme = manager.scheme[icd_scheme]
        assert isinstance(scheme, ICD10CM), f"Expected ICD10CM got {type(scheme)}."

        return CommonFlatCCS.register_flat_mappings(
            manager, "dx_flat_ccs", icd_scheme, cls.process_dx_ccs_icd_table(scheme)
        )

    @classmethod
    def register_pr_flat_ccs_maps(
        cls, manager: CodingSchemesManager, icd_scheme: ICD10PCSName
    ) -> CodingSchemesManager:  # expose
        scheme = manager.scheme[icd_scheme]
        assert isinstance(scheme, ICD10PCS), f"Expected PrFlatICD10, got {type(scheme)}."
        return CommonFlatCCS.register_flat_mappings(
            manager, "pr_flat_ccs", icd_scheme, cls.process_pr_ccs_icd_table(scheme)
        )

    @classmethod
    def register_dx_ccs_maps(
        cls, manager: CodingSchemesManager, icd_scheme: ICD10CMName
    ) -> CodingSchemesManager:  # expose
        scheme = manager.scheme[icd_scheme]
        assert isinstance(scheme, ICD10CM), f"Expected ICD10CM got {type(scheme)}."
        return cls.register_multi_level_mappings(manager, "dx_ccs", icd_scheme, cls.process_dx_ccs_icd_table(scheme))

    @classmethod
    def register_pr_ccs_maps(
        cls, manager: CodingSchemesManager, icd_scheme: ICD10PCSName
    ) -> CodingSchemesManager:  # expose
        scheme = manager.scheme[icd_scheme]
        assert isinstance(scheme, ICD10PCS), f"Expected PrFlatICD10, got {type(scheme)}."
        return cls.register_multi_level_mappings(manager, "pr_ccs", icd_scheme, cls.process_pr_ccs_icd_table(scheme))


class CCSMapRegistration:
    @staticmethod
    def icd9cm_maps(
        manager: CodingSchemesManager, icd9_scheme: ICD9CMName, dx_ccs: bool, dx_flat_ccs: bool
    ) -> CodingSchemesManager:
        if not any((dx_ccs, dx_flat_ccs)):
            return manager

        if dx_ccs:
            logging.debug(f"[BEGIN] mapping from {icd9_scheme} to dx_ccs")
            manager = MultiLevelCCSICD9MapOps.register_dx_ccs_maps(manager, icd9_scheme)
            logging.debug(f"[DONE] mapping from {icd9_scheme} to dx_ccs")
        if dx_flat_ccs:
            logging.debug(f"[BEGIN] mapping from {icd9_scheme} to dx_flat_ccs")
            manager = FlatCCS2ICD9MapOps.register_dx_flat_ccs_maps(manager, icd9_scheme)
            logging.debug(f"[DONE] mapping from {icd9_scheme} to dx_flat_ccs")
        return manager

    @staticmethod
    def icd9pcs_maps(
        manager: CodingSchemesManager, icd9_scheme: ICD9PCSName, pr_ccs: bool, pr_flat_ccs: bool
    ) -> CodingSchemesManager:
        if pr_ccs:
            logging.debug(f"[BEGIN] mapping from {icd9_scheme} to pr_ccs")
            manager = MultiLevelCCSICD9MapOps.register_pr_ccs_maps(manager, icd9_scheme)
            logging.debug(f"[DONE] mapping from {icd9_scheme} to pr_ccs")
        if pr_flat_ccs:
            logging.debug(f"[BEGIN] mapping from {icd9_scheme} to pr_flat_ccs")
            manager = FlatCCS2ICD9MapOps.register_pr_flat_ccs_maps(manager, icd9_scheme)
            logging.debug(f"[DONE] mapping from {icd9_scheme} to pr_flat_ccs")
        return manager

    @staticmethod
    def icd10cm_maps(
        manager: CodingSchemesManager, icd10_scheme: ICD10CMName, dx_ccs: bool, dx_flat_ccs: bool
    ) -> CodingSchemesManager:
        if dx_ccs:
            logging.debug(f"[BEGIN] mapping from {icd10_scheme} to dx_ccs")
            manager = CCS2ICD10MapOps.register_dx_ccs_maps(manager, icd10_scheme)
            logging.debug(f"[DONE] mapping from {icd10_scheme} to dx_ccs")
        if dx_flat_ccs:
            logging.debug(f"[BEGIN] mapping from {icd10_scheme} to dx_flat_ccs")
            manager = CCS2ICD10MapOps.register_dx_flat_ccs_maps(manager, icd10_scheme)
            logging.debug(f"[DONE] mapping from {icd10_scheme} to dx_flat_ccs")
        return manager

    @staticmethod
    def icd10pcs_maps(
        manager: CodingSchemesManager, icd10_scheme: ICD10PCSName, pr_ccs: bool, pr_flat_ccs: bool
    ) -> CodingSchemesManager:
        if pr_ccs:
            logging.debug(f"[BEGIN] mapping from {icd10_scheme} to pr_ccs")
            manager = CCS2ICD10MapOps.register_pr_ccs_maps(manager, icd10_scheme)
            logging.debug(f"[DONE] mapping from {icd10_scheme} to pr_ccs")
        if pr_flat_ccs:
            logging.debug(f"[BEGIN] mapping from {icd10_scheme} to pr_flat_ccs")
            manager = CCS2ICD10MapOps.register_pr_flat_ccs_maps(manager, icd10_scheme)
            logging.debug(f"[DONE] mapping from {icd10_scheme} to pr_flat_ccs")
        return manager

    @staticmethod
    def ccs_flat_to_multi_maps(
        manager: CodingSchemesManager, dx_bridge: ICD9CMName | None = None, pr_bridge: ICD9PCSName | None = None
    ):
        if dx_bridge is not None:
            logging.debug(f"[BEGIN] mapping from dx_flat_ccs to dx_ccs via {dx_bridge}")
            manager = manager.add_chained_map("dx_ccs", dx_bridge, "dx_flat_ccs")
            manager = manager.add_chained_map("dx_flat_ccs", dx_bridge, "dx_ccs")
            logging.debug(f"[DONE] mapping from dx_flat_ccs to dx_ccs via {dx_bridge}")
        if pr_bridge is not None:
            logging.debug(f"[BEGIN] mapping from pr_flat_ccs to pr_ccs via {pr_bridge}")
            manager = manager.add_chained_map("pr_ccs", pr_bridge, "pr_flat_ccs")
            manager = manager.add_chained_map("pr_flat_ccs", pr_bridge, "pr_ccs")
            logging.debug(f"[DONE] mapping from pr_flat_ccs to pr_ccs via {pr_bridge}")
        return manager
